"""Middleware plugins for agents."""

from .context_editing import (
    ClearToolUsesEdit,
    ContextEditingMiddleware,
)
from .human_in_the_loop import HumanInTheLoopMiddleware
from .model_call_limit import ModelCallLimitMiddleware
from .model_fallback import ModelFallbackMiddleware
from .pii import PIIDetectionError, PIIMiddleware
from .planning import PlanningMiddleware
from .prompt_caching import AnthropicPromptCachingMiddleware
from .summarization import SummarizationMiddleware
from .tool_call_limit import ToolCallLimitMiddleware
from .tool_selection import LLMToolSelectorMiddleware
from .types import (
    AgentMiddleware,
    AgentState,
    ModelRequest,
    after_agent,
    after_model,
    before_agent,
    before_model,
    dynamic_prompt,
    hook_config,
    modify_model_request,
)

__all__ = [
    "AgentMiddleware",
    "AgentState",
    # should move to langchain-anthropic if we decide to keep it
    "AnthropicPromptCachingMiddleware",
    "ClearToolUsesEdit",
    "ContextEditingMiddleware",
    "HumanInTheLoopMiddleware",
    "LLMToolSelectorMiddleware",
    "ModelCallLimitMiddleware",
    "ModelFallbackMiddleware",
    "ModelRequest",
    "PIIDetectionError",
    "PIIMiddleware",
    "PlanningMiddleware",
    "SummarizationMiddleware",
    "ToolCallLimitMiddleware",
    "after_agent",
    "after_model",
    "before_agent",
    "before_model",
    "dynamic_prompt",
    "hook_config",
    "modify_model_request",
]



"""Model fallback middleware for agents."""

from __future__ import annotations

from typing import TYPE_CHECKING

from langchain.agents.middleware.types import AgentMiddleware, AgentState, ModelRequest
from langchain.chat_models import init_chat_model

if TYPE_CHECKING:
    from langchain_core.language_models.chat_models import BaseChatModel
    from langgraph.runtime import Runtime


class ModelFallbackMiddleware(AgentMiddleware):
    """Middleware that provides automatic model fallback on errors.

    This middleware attempts to retry failed model calls with alternative models
    in sequence. When a model call fails, it tries the next model in the fallback
    list until either a call succeeds or all models have been exhausted.

    Example:
        ```python
        from langchain.agents.middleware.model_fallback import ModelFallbackMiddleware
        from langchain.agents import create_agent

        # Create middleware with fallback models (not including primary)
        fallback = ModelFallbackMiddleware(
            "openai:gpt-4o-mini",  # First fallback
            "anthropic:claude-3-5-sonnet-20241022",  # Second fallback
        )

        agent = create_agent(
            model="openai:gpt-4o",  # Primary model
            middleware=[fallback],
        )

        # If gpt-4o fails, automatically tries gpt-4o-mini, then claude
        result = await agent.invoke({"messages": [HumanMessage("Hello")]})
        ```
    """

    def __init__(
        self,
        first_model: str | BaseChatModel,
        *additional_models: str | BaseChatModel,
    ) -> None:
        """Initialize the model fallback middleware.

        Args:
            first_model: The first fallback model to try when the primary model fails.
                Can be a model name string or BaseChatModel instance.
            *additional_models: Additional fallback models to try, in order.
                Can be model name strings or BaseChatModel instances.
        """
        super().__init__()

        # Initialize all fallback models
        all_models = (first_model, *additional_models)
        self.models: list[BaseChatModel] = []
        for model in all_models:
            if isinstance(model, str):
                self.models.append(init_chat_model(model))
            else:
                self.models.append(model)

    def retry_model_request(
        self,
        error: Exception,  # noqa: ARG002
        request: ModelRequest,
        state: AgentState,  # noqa: ARG002
        runtime: Runtime,  # noqa: ARG002
        attempt: int,
    ) -> ModelRequest | None:
        """Retry with the next fallback model.

        Args:
            error: The exception that occurred during model invocation.
            request: The original model request that failed.
            state: The current agent state.
            runtime: The langgraph runtime.
            attempt: The current attempt number (1-indexed).

        Returns:
            ModelRequest with the next fallback model, or None if all models exhausted.
        """
        # attempt 1 = primary model failed, try models[0] (first fallback)
        fallback_index = attempt - 1
        # All fallback models exhausted
        if fallback_index >= len(self.models):
            return None
        # Try next fallback model
        request.model = self.models[fallback_index]
        return request



"""Call tracking middleware for agents."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from langchain_core.messages import AIMessage

from langchain.agents.middleware.types import AgentMiddleware, AgentState, hook_config

if TYPE_CHECKING:
    from langgraph.runtime import Runtime


def _build_limit_exceeded_message(
    thread_count: int,
    run_count: int,
    thread_limit: int | None,
    run_limit: int | None,
) -> str:
    """Build a message indicating which limits were exceeded.

    Args:
        thread_count: Current thread model call count.
        run_count: Current run model call count.
        thread_limit: Thread model call limit (if set).
        run_limit: Run model call limit (if set).

    Returns:
        A formatted message describing which limits were exceeded.
    """
    exceeded_limits = []
    if thread_limit is not None and thread_count >= thread_limit:
        exceeded_limits.append(f"thread limit ({thread_count}/{thread_limit})")
    if run_limit is not None and run_count >= run_limit:
        exceeded_limits.append(f"run limit ({run_count}/{run_limit})")

    return f"Model call limits exceeded: {', '.join(exceeded_limits)}"


class ModelCallLimitExceededError(Exception):
    """Exception raised when model call limits are exceeded.

    This exception is raised when the configured exit behavior is 'error'
    and either the thread or run model call limit has been exceeded.
    """

    def __init__(
        self,
        thread_count: int,
        run_count: int,
        thread_limit: int | None,
        run_limit: int | None,
    ) -> None:
        """Initialize the exception with call count information.

        Args:
            thread_count: Current thread model call count.
            run_count: Current run model call count.
            thread_limit: Thread model call limit (if set).
            run_limit: Run model call limit (if set).
        """
        self.thread_count = thread_count
        self.run_count = run_count
        self.thread_limit = thread_limit
        self.run_limit = run_limit

        msg = _build_limit_exceeded_message(thread_count, run_count, thread_limit, run_limit)
        super().__init__(msg)


class ModelCallLimitMiddleware(AgentMiddleware):
    """Middleware that tracks model call counts and enforces limits.

    This middleware monitors the number of model calls made during agent execution
    and can terminate the agent when specified limits are reached. It supports
    both thread-level and run-level call counting with configurable exit behaviors.

    Thread-level: The middleware tracks the number of model calls and persists
    call count across multiple runs (invocations) of the agent.

    Run-level: The middleware tracks the number of model calls made during a single
    run (invocation) of the agent.

    Example:
        ```python
        from langchain.agents.middleware.call_tracking import ModelCallLimitMiddleware
        from langchain.agents import create_agent

        # Create middleware with limits
        call_tracker = ModelCallLimitMiddleware(thread_limit=10, run_limit=5, exit_behavior="end")

        agent = create_agent("openai:gpt-4o", middleware=[call_tracker])

        # Agent will automatically jump to end when limits are exceeded
        result = await agent.invoke({"messages": [HumanMessage("Help me with a task")]})
        ```
    """

    def __init__(
        self,
        *,
        thread_limit: int | None = None,
        run_limit: int | None = None,
        exit_behavior: Literal["end", "error"] = "end",
    ) -> None:
        """Initialize the call tracking middleware.

        Args:
            thread_limit: Maximum number of model calls allowed per thread.
                None means no limit. Defaults to None.
            run_limit: Maximum number of model calls allowed per run.
                None means no limit. Defaults to None.
            exit_behavior: What to do when limits are exceeded.
                - "end": Jump to the end of the agent execution and
                    inject an artificial AI message indicating that the limit was exceeded.
                - "error": Raise a ModelCallLimitExceededError
                Defaults to "end".

        Raises:
            ValueError: If both limits are None or if exit_behavior is invalid.
        """
        super().__init__()

        if thread_limit is None and run_limit is None:
            msg = "At least one limit must be specified (thread_limit or run_limit)"
            raise ValueError(msg)

        if exit_behavior not in ("end", "error"):
            msg = f"Invalid exit_behavior: {exit_behavior}. Must be 'end' or 'error'"
            raise ValueError(msg)

        self.thread_limit = thread_limit
        self.run_limit = run_limit
        self.exit_behavior = exit_behavior

    @hook_config(can_jump_to=["end"])
    def before_model(self, state: AgentState, runtime: Runtime) -> dict[str, Any] | None:  # noqa: ARG002
        """Check model call limits before making a model call.

        Args:
            state: The current agent state containing call counts.
            runtime: The langgraph runtime.

        Returns:
            If limits are exceeded and exit_behavior is "end", returns
            a Command to jump to the end with a limit exceeded message. Otherwise returns None.

        Raises:
            ModelCallLimitExceededError: If limits are exceeded and exit_behavior
                is "error".
        """
        thread_count = state.get("thread_model_call_count", 0)
        run_count = state.get("run_model_call_count", 0)

        # Check if any limits will be exceeded after the next call
        thread_limit_exceeded = self.thread_limit is not None and thread_count >= self.thread_limit
        run_limit_exceeded = self.run_limit is not None and run_count >= self.run_limit

        if thread_limit_exceeded or run_limit_exceeded:
            if self.exit_behavior == "error":
                raise ModelCallLimitExceededError(
                    thread_count=thread_count,
                    run_count=run_count,
                    thread_limit=self.thread_limit,
                    run_limit=self.run_limit,
                )
            if self.exit_behavior == "end":
                # Create a message indicating the limit was exceeded
                limit_message = _build_limit_exceeded_message(
                    thread_count, run_count, self.thread_limit, self.run_limit
                )
                limit_ai_message = AIMessage(content=limit_message)

                return {"jump_to": "end", "messages": [limit_ai_message]}

        return None



"""Types for middleware and agents."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from inspect import iscoroutinefunction
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    Literal,
    Protocol,
    cast,
    overload,
)

from langchain_core.runnables import run_in_executor

if TYPE_CHECKING:
    from collections.abc import Awaitable

# needed as top level import for pydantic schema generation on AgentState
from langchain_core.messages import AnyMessage  # noqa: TC002
from langgraph.channels.ephemeral_value import EphemeralValue
from langgraph.channels.untracked_value import UntrackedValue
from langgraph.graph.message import add_messages
from langgraph.typing import ContextT
from typing_extensions import NotRequired, Required, TypedDict, TypeVar

if TYPE_CHECKING:
    from langchain_core.language_models.chat_models import BaseChatModel
    from langchain_core.tools import BaseTool
    from langgraph.runtime import Runtime
    from langgraph.types import Command

    from langchain.agents.structured_output import ResponseFormat

__all__ = [
    "AgentMiddleware",
    "AgentState",
    "ContextT",
    "ModelRequest",
    "OmitFromSchema",
    "PublicAgentState",
    "after_agent",
    "after_model",
    "before_agent",
    "before_model",
    "dynamic_prompt",
    "hook_config",
    "modify_model_request",
]

JumpTo = Literal["tools", "model", "end"]
"""Destination to jump to when a middleware node returns."""

ResponseT = TypeVar("ResponseT")


@dataclass
class ModelRequest:
    """Model request information for the agent."""

    model: BaseChatModel
    system_prompt: str | None
    messages: list[AnyMessage]  # excluding system prompt
    tool_choice: Any | None
    tools: list[BaseTool | dict]
    response_format: ResponseFormat | None
    model_settings: dict[str, Any] = field(default_factory=dict)


@dataclass
class OmitFromSchema:
    """Annotation used to mark state attributes as omitted from input or output schemas."""

    input: bool = True
    """Whether to omit the attribute from the input schema."""

    output: bool = True
    """Whether to omit the attribute from the output schema."""


OmitFromInput = OmitFromSchema(input=True, output=False)
"""Annotation used to mark state attributes as omitted from input schema."""

OmitFromOutput = OmitFromSchema(input=False, output=True)
"""Annotation used to mark state attributes as omitted from output schema."""

PrivateStateAttr = OmitFromSchema(input=True, output=True)
"""Annotation used to mark state attributes as purely internal for a given middleware."""


class AgentState(TypedDict, Generic[ResponseT]):
    """State schema for the agent."""

    messages: Required[Annotated[list[AnyMessage], add_messages]]
    jump_to: NotRequired[Annotated[JumpTo | None, EphemeralValue, PrivateStateAttr]]
    structured_response: NotRequired[Annotated[ResponseT, OmitFromInput]]
    thread_model_call_count: NotRequired[Annotated[int, PrivateStateAttr]]
    run_model_call_count: NotRequired[Annotated[int, UntrackedValue, PrivateStateAttr]]


class PublicAgentState(TypedDict, Generic[ResponseT]):
    """Public state schema for the agent.

    Just used for typing purposes.
    """

    messages: Required[Annotated[list[AnyMessage], add_messages]]
    structured_response: NotRequired[ResponseT]


StateT = TypeVar("StateT", bound=AgentState, default=AgentState)
StateT_contra = TypeVar("StateT_contra", bound=AgentState, contravariant=True)


class AgentMiddleware(Generic[StateT, ContextT]):
    """Base middleware class for an agent.

    Subclass this and implement any of the defined methods to customize agent behavior
    between steps in the main agent loop.
    """

    state_schema: type[StateT] = cast("type[StateT]", AgentState)
    """The schema for state passed to the middleware nodes."""

    tools: list[BaseTool]
    """Additional tools registered by the middleware."""

    @property
    def name(self) -> str:
        """The name of the middleware instance.

        Defaults to the class name, but can be overridden for custom naming.
        """
        return self.__class__.__name__

    def before_agent(self, state: StateT, runtime: Runtime[ContextT]) -> dict[str, Any] | None:
        """Logic to run before the agent execution starts."""

    async def abefore_agent(
        self, state: StateT, runtime: Runtime[ContextT]
    ) -> dict[str, Any] | None:
        """Async logic to run before the agent execution starts."""

    def before_model(self, state: StateT, runtime: Runtime[ContextT]) -> dict[str, Any] | None:
        """Logic to run before the model is called."""

    async def abefore_model(
        self, state: StateT, runtime: Runtime[ContextT]
    ) -> dict[str, Any] | None:
        """Async logic to run before the model is called."""

    def modify_model_request(
        self,
        request: ModelRequest,
        state: StateT,  # noqa: ARG002
        runtime: Runtime[ContextT],  # noqa: ARG002
    ) -> ModelRequest:
        """Logic to modify request kwargs before the model is called."""
        return request

    async def amodify_model_request(
        self,
        request: ModelRequest,
        state: StateT,
        runtime: Runtime[ContextT],
    ) -> ModelRequest:
        """Async logic to modify request kwargs before the model is called."""
        return await run_in_executor(None, self.modify_model_request, request, state, runtime)

    def after_model(self, state: StateT, runtime: Runtime[ContextT]) -> dict[str, Any] | None:
        """Logic to run after the model is called."""

    async def aafter_model(
        self, state: StateT, runtime: Runtime[ContextT]
    ) -> dict[str, Any] | None:
        """Async logic to run after the model is called."""

    def retry_model_request(
        self,
        error: Exception,  # noqa: ARG002
        request: ModelRequest,  # noqa: ARG002
        state: StateT,  # noqa: ARG002
        runtime: Runtime[ContextT],  # noqa: ARG002
        attempt: int,  # noqa: ARG002
    ) -> ModelRequest | None:
        """Logic to handle model invocation errors and optionally retry.

        Args:
            error: The exception that occurred during model invocation.
            request: The original model request that failed.
            state: The current agent state.
            runtime: The langgraph runtime.
            attempt: The current attempt number (1-indexed).

        Returns:
            ModelRequest: Modified request to retry with.
            None: Propagate the error (re-raise).
        """
        return None

    async def aretry_model_request(
        self,
        error: Exception,
        request: ModelRequest,
        state: StateT,
        runtime: Runtime[ContextT],
        attempt: int,
    ) -> ModelRequest | None:
        """Async logic to handle model invocation errors and optionally retry.

        Args:
            error: The exception that occurred during model invocation.
            request: The original model request that failed.
            state: The current agent state.
            runtime: The langgraph runtime.
            attempt: The current attempt number (1-indexed).

        Returns:
            ModelRequest: Modified request to retry with.
            None: Propagate the error (re-raise).
        """
        return await run_in_executor(
            None, self.retry_model_request, error, request, state, runtime, attempt
        )

    def after_agent(self, state: StateT, runtime: Runtime[ContextT]) -> dict[str, Any] | None:
        """Logic to run after the agent execution completes."""

    async def aafter_agent(
        self, state: StateT, runtime: Runtime[ContextT]
    ) -> dict[str, Any] | None:
        """Async logic to run after the agent execution completes."""


class _CallableWithStateAndRuntime(Protocol[StateT_contra, ContextT]):
    """Callable with AgentState and Runtime as arguments."""

    def __call__(
        self, state: StateT_contra, runtime: Runtime[ContextT]
    ) -> dict[str, Any] | Command | None | Awaitable[dict[str, Any] | Command | None]:
        """Perform some logic with the state and runtime."""
        ...


class _CallableWithModelRequestAndStateAndRuntime(Protocol[StateT_contra, ContextT]):
    """Callable with ModelRequest, AgentState, and Runtime as arguments."""

    def __call__(
        self, request: ModelRequest, state: StateT_contra, runtime: Runtime[ContextT]
    ) -> ModelRequest | Awaitable[ModelRequest]:
        """Perform some logic with the model request, state, and runtime."""
        ...


class _CallableReturningPromptString(Protocol[StateT_contra, ContextT]):
    """Callable that returns a prompt string given ModelRequest, AgentState, and Runtime."""

    def __call__(
        self, request: ModelRequest, state: StateT_contra, runtime: Runtime[ContextT]
    ) -> str | Awaitable[str]:
        """Generate a system prompt string based on the request, state, and runtime."""
        ...


CallableT = TypeVar("CallableT", bound=Callable[..., Any])


def hook_config(
    *,
    can_jump_to: list[JumpTo] | None = None,
) -> Callable[[CallableT], CallableT]:
    """Decorator to configure hook behavior in middleware methods.

    Use this decorator on `before_model` or `after_model` methods in middleware classes
    to configure their behavior. Currently supports specifying which destinations they
    can jump to, which establishes conditional edges in the agent graph.

    Args:
        can_jump_to: Optional list of valid jump destinations. Can be:
            - "tools": Jump to the tools node
            - "model": Jump back to the model node
            - "end": Jump to the end of the graph

    Returns:
        Decorator function that marks the method with configuration metadata.

    Examples:
        Using decorator on a class method:
        ```python
        class MyMiddleware(AgentMiddleware):
            @hook_config(can_jump_to=["end", "model"])
            def before_model(self, state: AgentState) -> dict[str, Any] | None:
                if some_condition(state):
                    return {"jump_to": "end"}
                return None
        ```

        Alternative: Use the `can_jump_to` parameter in `before_model`/`after_model` decorators:
        ```python
        @before_model(can_jump_to=["end"])
        def conditional_middleware(state: AgentState) -> dict[str, Any] | None:
            if should_exit(state):
                return {"jump_to": "end"}
            return None
        ```
    """

    def decorator(func: CallableT) -> CallableT:
        if can_jump_to is not None:
            func.__can_jump_to__ = can_jump_to  # type: ignore[attr-defined]
        return func

    return decorator


@overload
def before_model(
    func: _CallableWithStateAndRuntime[StateT, ContextT],
) -> AgentMiddleware[StateT, ContextT]: ...


@overload
def before_model(
    func: None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> Callable[
    [_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]
]: ...


def before_model(
    func: _CallableWithStateAndRuntime[StateT, ContextT] | None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> (
    Callable[[_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]]
    | AgentMiddleware[StateT, ContextT]
):
    """Decorator used to dynamically create a middleware with the before_model hook.

    Args:
        func: The function to be decorated. Must accept:
            `state: StateT, runtime: Runtime[ContextT]` - State and runtime context
        state_schema: Optional custom state schema type. If not provided, uses the default
            AgentState schema.
        tools: Optional list of additional tools to register with this middleware.
        can_jump_to: Optional list of valid jump destinations for conditional edges.
            Valid values are: "tools", "model", "end"
        name: Optional name for the generated middleware class. If not provided,
            uses the decorated function's name.

    Returns:
        Either an AgentMiddleware instance (if func is provided directly) or a decorator function
        that can be applied to a function its wrapping.

    The decorated function should return:
        - `dict[str, Any]` - State updates to merge into the agent state
        - `Command` - A command to control flow (e.g., jump to different node)
        - `None` - No state updates or flow control

    Examples:
        Basic usage:
        ```python
        @before_model
        def log_before_model(state: AgentState, runtime: Runtime) -> None:
            print(f"About to call model with {len(state['messages'])} messages")
        ```

        With conditional jumping:
        ```python
        @before_model(can_jump_to=["end"])
        def conditional_before_model(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
            if some_condition(state):
                return {"jump_to": "end"}
            return None
        ```

        With custom state schema:
        ```python
        @before_model(state_schema=MyCustomState)
        def custom_before_model(state: MyCustomState, runtime: Runtime) -> dict[str, Any]:
            return {"custom_field": "updated_value"}
        ```
    """

    def decorator(
        func: _CallableWithStateAndRuntime[StateT, ContextT],
    ) -> AgentMiddleware[StateT, ContextT]:
        is_async = iscoroutinefunction(func)

        func_can_jump_to = (
            can_jump_to if can_jump_to is not None else getattr(func, "__can_jump_to__", [])
        )

        if is_async:

            async def async_wrapped(
                self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
                state: StateT,
                runtime: Runtime[ContextT],
            ) -> dict[str, Any] | Command | None:
                return await func(state, runtime)  # type: ignore[misc]

            # Preserve can_jump_to metadata on the wrapped function
            if func_can_jump_to:
                async_wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

            middleware_name = name or cast(
                "str", getattr(func, "__name__", "BeforeModelMiddleware")
            )

            return type(
                middleware_name,
                (AgentMiddleware,),
                {
                    "state_schema": state_schema or AgentState,
                    "tools": tools or [],
                    "abefore_model": async_wrapped,
                },
            )()

        def wrapped(
            self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
            state: StateT,
            runtime: Runtime[ContextT],
        ) -> dict[str, Any] | Command | None:
            return func(state, runtime)  # type: ignore[return-value]

        # Preserve can_jump_to metadata on the wrapped function
        if func_can_jump_to:
            wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

        # Use function name as default if no name provided
        middleware_name = name or cast("str", getattr(func, "__name__", "BeforeModelMiddleware"))

        return type(
            middleware_name,
            (AgentMiddleware,),
            {
                "state_schema": state_schema or AgentState,
                "tools": tools or [],
                "before_model": wrapped,
            },
        )()

    if func is not None:
        return decorator(func)
    return decorator


@overload
def modify_model_request(
    func: _CallableWithModelRequestAndStateAndRuntime[StateT, ContextT],
) -> AgentMiddleware[StateT, ContextT]: ...


@overload
def modify_model_request(
    func: None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    name: str | None = None,
) -> Callable[
    [_CallableWithModelRequestAndStateAndRuntime[StateT, ContextT]],
    AgentMiddleware[StateT, ContextT],
]: ...


def modify_model_request(
    func: _CallableWithModelRequestAndStateAndRuntime[StateT, ContextT] | None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    name: str | None = None,
) -> (
    Callable[
        [_CallableWithModelRequestAndStateAndRuntime[StateT, ContextT]],
        AgentMiddleware[StateT, ContextT],
    ]
    | AgentMiddleware[StateT, ContextT]
):
    r"""Decorator used to dynamically create a middleware with the modify_model_request hook.

    Args:
        func: The function to be decorated. Must accept:
            `request: ModelRequest, state: StateT, runtime: Runtime[ContextT]` -
            Model request, state, and runtime context
        state_schema: Optional custom state schema type. If not provided, uses the default
            AgentState schema.
        tools: Optional list of additional tools to register with this middleware.
        name: Optional name for the generated middleware class. If not provided,
            uses the decorated function's name.

    Returns:
        Either an AgentMiddleware instance (if func is provided) or a decorator function
        that can be applied to a function.

    The decorated function should return:
        - `ModelRequest` - The modified model request to be sent to the language model

    Examples:
        Basic usage to modify system prompt:
        ```python
        @modify_model_request
        def add_context_to_prompt(
            request: ModelRequest, state: AgentState, runtime: Runtime
        ) -> ModelRequest:
            if request.system_prompt:
                request.system_prompt += "\n\nAdditional context: ..."
            else:
                request.system_prompt = "Additional context: ..."
            return request
        ```

        Usage with runtime and custom model settings:
        ```python
        @modify_model_request
        def dynamic_model_settings(
            request: ModelRequest, state: AgentState, runtime: Runtime
        ) -> ModelRequest:
            # Use a different model based on user subscription tier
            if runtime.context.get("subscription_tier") == "premium":
                request.model = "gpt-4o"
            else:
                request.model = "gpt-4o-mini"

            return request
        ```
    """

    def decorator(
        func: _CallableWithModelRequestAndStateAndRuntime[StateT, ContextT],
    ) -> AgentMiddleware[StateT, ContextT]:
        is_async = iscoroutinefunction(func)

        if is_async:

            async def async_wrapped(
                self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
                request: ModelRequest,
                state: StateT,
                runtime: Runtime[ContextT],
            ) -> ModelRequest:
                return await func(request, state, runtime)  # type: ignore[misc]

            middleware_name = name or cast(
                "str", getattr(func, "__name__", "ModifyModelRequestMiddleware")
            )

            return type(
                middleware_name,
                (AgentMiddleware,),
                {
                    "state_schema": state_schema or AgentState,
                    "tools": tools or [],
                    "amodify_model_request": async_wrapped,
                },
            )()

        def wrapped(
            self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
            request: ModelRequest,
            state: StateT,
            runtime: Runtime[ContextT],
        ) -> ModelRequest:
            return func(request, state, runtime)  # type: ignore[return-value]

        middleware_name = name or cast(
            "str", getattr(func, "__name__", "ModifyModelRequestMiddleware")
        )

        return type(
            middleware_name,
            (AgentMiddleware,),
            {
                "state_schema": state_schema or AgentState,
                "tools": tools or [],
                "modify_model_request": wrapped,
            },
        )()

    if func is not None:
        return decorator(func)
    return decorator


@overload
def after_model(
    func: _CallableWithStateAndRuntime[StateT, ContextT],
) -> AgentMiddleware[StateT, ContextT]: ...


@overload
def after_model(
    func: None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> Callable[
    [_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]
]: ...


def after_model(
    func: _CallableWithStateAndRuntime[StateT, ContextT] | None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> (
    Callable[[_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]]
    | AgentMiddleware[StateT, ContextT]
):
    """Decorator used to dynamically create a middleware with the after_model hook.

    Args:
        func: The function to be decorated. Must accept:
            `state: StateT, runtime: Runtime[ContextT]` - State and runtime context
        state_schema: Optional custom state schema type. If not provided, uses the default
            AgentState schema.
        tools: Optional list of additional tools to register with this middleware.
        can_jump_to: Optional list of valid jump destinations for conditional edges.
            Valid values are: "tools", "model", "end"
        name: Optional name for the generated middleware class. If not provided,
            uses the decorated function's name.

    Returns:
        Either an AgentMiddleware instance (if func is provided) or a decorator function
        that can be applied to a function.

    The decorated function should return:
        - `dict[str, Any]` - State updates to merge into the agent state
        - `Command` - A command to control flow (e.g., jump to different node)
        - `None` - No state updates or flow control

    Examples:
        Basic usage for logging model responses:
        ```python
        @after_model
        def log_latest_message(state: AgentState, runtime: Runtime) -> None:
            print(state["messages"][-1].content)
        ```

        With custom state schema:
        ```python
        @after_model(state_schema=MyCustomState, name="MyAfterModelMiddleware")
        def custom_after_model(state: MyCustomState, runtime: Runtime) -> dict[str, Any]:
            return {"custom_field": "updated_after_model"}
        ```
    """

    def decorator(
        func: _CallableWithStateAndRuntime[StateT, ContextT],
    ) -> AgentMiddleware[StateT, ContextT]:
        is_async = iscoroutinefunction(func)
        # Extract can_jump_to from decorator parameter or from function metadata
        func_can_jump_to = (
            can_jump_to if can_jump_to is not None else getattr(func, "__can_jump_to__", [])
        )

        if is_async:

            async def async_wrapped(
                self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
                state: StateT,
                runtime: Runtime[ContextT],
            ) -> dict[str, Any] | Command | None:
                return await func(state, runtime)  # type: ignore[misc]

            # Preserve can_jump_to metadata on the wrapped function
            if func_can_jump_to:
                async_wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

            middleware_name = name or cast("str", getattr(func, "__name__", "AfterModelMiddleware"))

            return type(
                middleware_name,
                (AgentMiddleware,),
                {
                    "state_schema": state_schema or AgentState,
                    "tools": tools or [],
                    "aafter_model": async_wrapped,
                },
            )()

        def wrapped(
            self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
            state: StateT,
            runtime: Runtime[ContextT],
        ) -> dict[str, Any] | Command | None:
            return func(state, runtime)  # type: ignore[return-value]

        # Preserve can_jump_to metadata on the wrapped function
        if func_can_jump_to:
            wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

        # Use function name as default if no name provided
        middleware_name = name or cast("str", getattr(func, "__name__", "AfterModelMiddleware"))

        return type(
            middleware_name,
            (AgentMiddleware,),
            {
                "state_schema": state_schema or AgentState,
                "tools": tools or [],
                "after_model": wrapped,
            },
        )()

    if func is not None:
        return decorator(func)
    return decorator


@overload
def before_agent(
    func: _CallableWithStateAndRuntime[StateT, ContextT],
) -> AgentMiddleware[StateT, ContextT]: ...


@overload
def before_agent(
    func: None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> Callable[
    [_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]
]: ...


def before_agent(
    func: _CallableWithStateAndRuntime[StateT, ContextT] | None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> (
    Callable[[_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]]
    | AgentMiddleware[StateT, ContextT]
):
    """Decorator used to dynamically create a middleware with the before_agent hook.

    Args:
        func: The function to be decorated. Must accept:
            `state: StateT, runtime: Runtime[ContextT]` - State and runtime context
        state_schema: Optional custom state schema type. If not provided, uses the default
            AgentState schema.
        tools: Optional list of additional tools to register with this middleware.
        can_jump_to: Optional list of valid jump destinations for conditional edges.
            Valid values are: "tools", "model", "end"
        name: Optional name for the generated middleware class. If not provided,
            uses the decorated function's name.

    Returns:
        Either an AgentMiddleware instance (if func is provided directly) or a decorator function
        that can be applied to a function its wrapping.

    The decorated function should return:
        - `dict[str, Any]` - State updates to merge into the agent state
        - `Command` - A command to control flow (e.g., jump to different node)
        - `None` - No state updates or flow control

    Examples:
        Basic usage:
        ```python
        @before_agent
        def log_before_agent(state: AgentState, runtime: Runtime) -> None:
            print(f"Starting agent with {len(state['messages'])} messages")
        ```

        With conditional jumping:
        ```python
        @before_agent(can_jump_to=["end"])
        def conditional_before_agent(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
            if some_condition(state):
                return {"jump_to": "end"}
            return None
        ```

        With custom state schema:
        ```python
        @before_agent(state_schema=MyCustomState)
        def custom_before_agent(state: MyCustomState, runtime: Runtime) -> dict[str, Any]:
            return {"custom_field": "initialized_value"}
        ```
    """

    def decorator(
        func: _CallableWithStateAndRuntime[StateT, ContextT],
    ) -> AgentMiddleware[StateT, ContextT]:
        is_async = iscoroutinefunction(func)

        func_can_jump_to = (
            can_jump_to if can_jump_to is not None else getattr(func, "__can_jump_to__", [])
        )

        if is_async:

            async def async_wrapped(
                self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
                state: StateT,
                runtime: Runtime[ContextT],
            ) -> dict[str, Any] | Command | None:
                return await func(state, runtime)  # type: ignore[misc]

            # Preserve can_jump_to metadata on the wrapped function
            if func_can_jump_to:
                async_wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

            middleware_name = name or cast(
                "str", getattr(func, "__name__", "BeforeAgentMiddleware")
            )

            return type(
                middleware_name,
                (AgentMiddleware,),
                {
                    "state_schema": state_schema or AgentState,
                    "tools": tools or [],
                    "abefore_agent": async_wrapped,
                },
            )()

        def wrapped(
            self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
            state: StateT,
            runtime: Runtime[ContextT],
        ) -> dict[str, Any] | Command | None:
            return func(state, runtime)  # type: ignore[return-value]

        # Preserve can_jump_to metadata on the wrapped function
        if func_can_jump_to:
            wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

        # Use function name as default if no name provided
        middleware_name = name or cast("str", getattr(func, "__name__", "BeforeAgentMiddleware"))

        return type(
            middleware_name,
            (AgentMiddleware,),
            {
                "state_schema": state_schema or AgentState,
                "tools": tools or [],
                "before_agent": wrapped,
            },
        )()

    if func is not None:
        return decorator(func)
    return decorator


@overload
def after_agent(
    func: _CallableWithStateAndRuntime[StateT, ContextT],
) -> AgentMiddleware[StateT, ContextT]: ...


@overload
def after_agent(
    func: None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> Callable[
    [_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]
]: ...


def after_agent(
    func: _CallableWithStateAndRuntime[StateT, ContextT] | None = None,
    *,
    state_schema: type[StateT] | None = None,
    tools: list[BaseTool] | None = None,
    can_jump_to: list[JumpTo] | None = None,
    name: str | None = None,
) -> (
    Callable[[_CallableWithStateAndRuntime[StateT, ContextT]], AgentMiddleware[StateT, ContextT]]
    | AgentMiddleware[StateT, ContextT]
):
    """Decorator used to dynamically create a middleware with the after_agent hook.

    Args:
        func: The function to be decorated. Must accept:
            `state: StateT, runtime: Runtime[ContextT]` - State and runtime context
        state_schema: Optional custom state schema type. If not provided, uses the default
            AgentState schema.
        tools: Optional list of additional tools to register with this middleware.
        can_jump_to: Optional list of valid jump destinations for conditional edges.
            Valid values are: "tools", "model", "end"
        name: Optional name for the generated middleware class. If not provided,
            uses the decorated function's name.

    Returns:
        Either an AgentMiddleware instance (if func is provided) or a decorator function
        that can be applied to a function.

    The decorated function should return:
        - `dict[str, Any]` - State updates to merge into the agent state
        - `Command` - A command to control flow (e.g., jump to different node)
        - `None` - No state updates or flow control

    Examples:
        Basic usage for logging agent completion:
        ```python
        @after_agent
        def log_completion(state: AgentState, runtime: Runtime) -> None:
            print(f"Agent completed with {len(state['messages'])} messages")
        ```

        With custom state schema:
        ```python
        @after_agent(state_schema=MyCustomState, name="MyAfterAgentMiddleware")
        def custom_after_agent(state: MyCustomState, runtime: Runtime) -> dict[str, Any]:
            return {"custom_field": "finalized_value"}
        ```
    """

    def decorator(
        func: _CallableWithStateAndRuntime[StateT, ContextT],
    ) -> AgentMiddleware[StateT, ContextT]:
        is_async = iscoroutinefunction(func)
        # Extract can_jump_to from decorator parameter or from function metadata
        func_can_jump_to = (
            can_jump_to if can_jump_to is not None else getattr(func, "__can_jump_to__", [])
        )

        if is_async:

            async def async_wrapped(
                self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
                state: StateT,
                runtime: Runtime[ContextT],
            ) -> dict[str, Any] | Command | None:
                return await func(state, runtime)  # type: ignore[misc]

            # Preserve can_jump_to metadata on the wrapped function
            if func_can_jump_to:
                async_wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

            middleware_name = name or cast("str", getattr(func, "__name__", "AfterAgentMiddleware"))

            return type(
                middleware_name,
                (AgentMiddleware,),
                {
                    "state_schema": state_schema or AgentState,
                    "tools": tools or [],
                    "aafter_agent": async_wrapped,
                },
            )()

        def wrapped(
            self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
            state: StateT,
            runtime: Runtime[ContextT],
        ) -> dict[str, Any] | Command | None:
            return func(state, runtime)  # type: ignore[return-value]

        # Preserve can_jump_to metadata on the wrapped function
        if func_can_jump_to:
            wrapped.__can_jump_to__ = func_can_jump_to  # type: ignore[attr-defined]

        # Use function name as default if no name provided
        middleware_name = name or cast("str", getattr(func, "__name__", "AfterAgentMiddleware"))

        return type(
            middleware_name,
            (AgentMiddleware,),
            {
                "state_schema": state_schema or AgentState,
                "tools": tools or [],
                "after_agent": wrapped,
            },
        )()

    if func is not None:
        return decorator(func)
    return decorator


@overload
def dynamic_prompt(
    func: _CallableReturningPromptString[StateT, ContextT],
) -> AgentMiddleware[StateT, ContextT]: ...


@overload
def dynamic_prompt(
    func: None = None,
) -> Callable[
    [_CallableReturningPromptString[StateT, ContextT]],
    AgentMiddleware[StateT, ContextT],
]: ...


def dynamic_prompt(
    func: _CallableReturningPromptString[StateT, ContextT] | None = None,
) -> (
    Callable[
        [_CallableReturningPromptString[StateT, ContextT]],
        AgentMiddleware[StateT, ContextT],
    ]
    | AgentMiddleware[StateT, ContextT]
):
    """Decorator used to dynamically generate system prompts for the model.

    This is a convenience decorator that creates middleware using `modify_model_request`
    specifically for dynamic prompt generation. The decorated function should return
    a string that will be set as the system prompt for the model request.

    Args:
        func: The function to be decorated. Must accept:
            `request: ModelRequest, state: StateT, runtime: Runtime[ContextT]` -
            Model request, state, and runtime context

    Returns:
        Either an AgentMiddleware instance (if func is provided) or a decorator function
        that can be applied to a function.

    The decorated function should return:
        - `str` - The system prompt to use for the model request

    Examples:
        Basic usage with dynamic content:
        ```python
        @dynamic_prompt
        def my_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
            user_name = runtime.context.get("user_name", "User")
            return f"You are a helpful assistant helping {user_name}."
        ```

        Using state to customize the prompt:
        ```python
        @dynamic_prompt
        def context_aware_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
            msg_count = len(state["messages"])
            if msg_count > 10:
                return "You are in a long conversation. Be concise."
            return "You are a helpful assistant."
        ```

        Using with agent:
        ```python
        agent = create_agent(model, middleware=[my_prompt])
        ```
    """

    def decorator(
        func: _CallableReturningPromptString[StateT, ContextT],
    ) -> AgentMiddleware[StateT, ContextT]:
        is_async = iscoroutinefunction(func)

        if is_async:

            async def async_wrapped(
                self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
                request: ModelRequest,
                state: StateT,
                runtime: Runtime[ContextT],
            ) -> ModelRequest:
                prompt = await func(request, state, runtime)  # type: ignore[misc]
                request.system_prompt = prompt
                return request

            middleware_name = cast("str", getattr(func, "__name__", "DynamicPromptMiddleware"))

            return type(
                middleware_name,
                (AgentMiddleware,),
                {
                    "state_schema": AgentState,
                    "tools": [],
                    "amodify_model_request": async_wrapped,
                },
            )()

        def wrapped(
            self: AgentMiddleware[StateT, ContextT],  # noqa: ARG001
            request: ModelRequest,
            state: StateT,
            runtime: Runtime[ContextT],
        ) -> ModelRequest:
            prompt = cast("str", func(request, state, runtime))
            request.system_prompt = prompt
            return request

        middleware_name = cast("str", getattr(func, "__name__", "DynamicPromptMiddleware"))

        return type(
            middleware_name,
            (AgentMiddleware,),
            {
                "state_schema": AgentState,
                "tools": [],
                "modify_model_request": wrapped,
            },
        )()

    if func is not None:
        return decorator(func)
    return decorator



# import dataclasses
# import inspect
# from types import UnionType
# from typing import (
#     Annotated,
#     Union,
# )

# import pytest
# from langchain_core.language_models import BaseChatModel
# from langchain_core.messages import (
#     AIMessage,
#     HumanMessage,
#     MessageLikeRepresentation,
#     RemoveMessage,
#     SystemMessage,
#     ToolCall,
#     ToolMessage,
# )
# from langchain_core.runnables import RunnableConfig, RunnableLambda
# from langchain_core.tools import BaseTool, InjectedToolCallId, ToolException
# from langchain_core.tools import tool as dec_tool
# from langgraph.checkpoint.base import BaseCheckpointSaver
# from langgraph.graph import START, MessagesState, StateGraph
# from langgraph.graph.message import REMOVE_ALL_MESSAGES
# from langgraph.runtime import Runtime
# from langgraph.store.base import BaseStore
# from langgraph.store.memory import InMemoryStore
# from langgraph.types import Command, Interrupt, interrupt
# from pydantic import BaseModel, Field
# from typing_extensions import TypedDict

# from langchain.agents import (
#     AgentState,
#     create_agent,
# )
# from langchain.tools import (
#     ToolNode,
#     InjectedState,
#     InjectedStore,
# )
# from langchain.tools.tool_node import (
#     _get_state_args,
#     _infer_handled_types,
# )

# from .any_str import AnyStr
# from .messages import _AnyIdHumanMessage, _AnyIdToolMessage
# from .model import FakeToolCallingModel

# pytestmark = pytest.mark.anyio


# def test_no_prompt(sync_checkpointer: BaseCheckpointSaver) -> None:
#     model = FakeToolCallingModel()

#     agent = create_agent(
#         model,
#         [],
#         checkpointer=sync_checkpointer,
#     )
#     inputs = [HumanMessage("hi?")]
#     thread = {"configurable": {"thread_id": "123"}}
#     response = agent.invoke({"messages": inputs}, thread, debug=True)
#     expected_response = {"messages": [*inputs, AIMessage(content="hi?", id="0")]}
#     assert response == expected_response

#     saved = sync_checkpointer.get_tuple(thread)
#     assert saved is not None
#     assert saved.checkpoint["channel_values"] == {
#         "messages": [
#             _AnyIdHumanMessage(content="hi?"),
#             AIMessage(content="hi?", id="0"),
#         ],
#     }
#     assert saved.metadata == {
#         "parents": {},
#         "source": "loop",
#         "step": 1,
#     }
#     assert saved.pending_writes == []


# async def test_no_prompt_async(async_checkpointer: BaseCheckpointSaver) -> None:
#     model = FakeToolCallingModel()

#     agent = create_agent(model, [], checkpointer=async_checkpointer)
#     inputs = [HumanMessage("hi?")]
#     thread = {"configurable": {"thread_id": "123"}}
#     response = await agent.ainvoke({"messages": inputs}, thread, debug=True)
#     expected_response = {"messages": [*inputs, AIMessage(content="hi?", id="0")]}
#     assert response == expected_response

#     saved = await async_checkpointer.aget_tuple(thread)
#     assert saved is not None
#     assert saved.checkpoint["channel_values"] == {
#         "messages": [
#             _AnyIdHumanMessage(content="hi?"),
#             AIMessage(content="hi?", id="0"),
#         ],
#     }
#     assert saved.metadata == {
#         "parents": {},
#         "source": "loop",
#         "step": 1,
#     }
#     assert saved.pending_writes == []


# def test_system_message_prompt() -> None:
#     prompt = SystemMessage(content="Foo")
#     agent = create_agent(FakeToolCallingModel(), [], system_prompt=prompt)
#     inputs = [HumanMessage("hi?")]
#     response = agent.invoke({"messages": inputs})
#     expected_response = {"messages": [*inputs, AIMessage(content="Foo-hi?", id="0", tool_calls=[])]}
#     assert response == expected_response


# def test_string_prompt() -> None:
#     prompt = "Foo"
#     agent = create_agent(FakeToolCallingModel(), [], system_prompt=prompt)
#     inputs = [HumanMessage("hi?")]
#     response = agent.invoke({"messages": inputs})
#     expected_response = {"messages": [*inputs, AIMessage(content="Foo-hi?", id="0", tool_calls=[])]}
#     assert response == expected_response


# def test_callable_prompt() -> None:
#     def prompt(state):
#         modified_message = f"Bar {state['messages'][-1].content}"
#         return [HumanMessage(content=modified_message)]

#     agent = create_agent(FakeToolCallingModel(), [], system_prompt=prompt)
#     inputs = [HumanMessage("hi?")]
#     response = agent.invoke({"messages": inputs})
#     expected_response = {"messages": [*inputs, AIMessage(content="Bar hi?", id="0")]}
#     assert response == expected_response


# async def test_callable_prompt_async() -> None:
#     async def prompt(state):
#         modified_message = f"Bar {state['messages'][-1].content}"
#         return [HumanMessage(content=modified_message)]

#     agent = create_agent(FakeToolCallingModel(), [], system_prompt=prompt)
#     inputs = [HumanMessage("hi?")]
#     response = await agent.ainvoke({"messages": inputs})
#     expected_response = {"messages": [*inputs, AIMessage(content="Bar hi?", id="0")]}
#     assert response == expected_response


# def test_runnable_prompt() -> None:
#     prompt = RunnableLambda(
#         lambda state: [HumanMessage(content=f"Baz {state['messages'][-1].content}")]
#     )

#     agent = create_agent(FakeToolCallingModel(), [], system_prompt=prompt)
#     inputs = [HumanMessage("hi?")]
#     response = agent.invoke({"messages": inputs})
#     expected_response = {"messages": [*inputs, AIMessage(content="Baz hi?", id="0")]}
#     assert response == expected_response


# def test_prompt_with_store() -> None:
#     def add(a: int, b: int):
#         """Adds a and b"""
#         return a + b

#     in_memory_store = InMemoryStore()
#     in_memory_store.put(("memories", "1"), "user_name", {"data": "User name is Alice"})
#     in_memory_store.put(("memories", "2"), "user_name", {"data": "User name is Bob"})

#     def prompt(state, config, *, store):
#         user_id = config["configurable"]["user_id"]
#         system_str = store.get(("memories", user_id), "user_name").value["data"]
#         return [SystemMessage(system_str)] + state["messages"]

#     def prompt_no_store(state, config):
#         return SystemMessage("foo") + state["messages"]

#     model = FakeToolCallingModel()

#     # test state modifier that uses store works
#     agent = create_agent(
#         model,
#         [add],
#         prompt=prompt,
#         store=in_memory_store,
#     )
#     response = agent.invoke({"messages": [("user", "hi")]}, {"configurable": {"user_id": "1"}})
#     assert response["messages"][-1].content == "User name is Alice-hi"

#     # test state modifier that doesn't use store works
#     agent = create_agent(
#         model,
#         [add],
#         prompt=prompt_no_store,
#         store=in_memory_store,
#     )
#     response = agent.invoke({"messages": [("user", "hi")]}, {"configurable": {"user_id": "2"}})
#     assert response["messages"][-1].content == "foo-hi"


# async def test_prompt_with_store_async() -> None:
#     async def add(a: int, b: int):
#         """Adds a and b"""
#         return a + b

#     in_memory_store = InMemoryStore()
#     await in_memory_store.aput(("memories", "1"), "user_name", {"data": "User name is Alice"})
#     await in_memory_store.aput(("memories", "2"), "user_name", {"data": "User name is Bob"})

#     async def prompt(state, config, *, store):
#         user_id = config["configurable"]["user_id"]
#         system_str = (await store.aget(("memories", user_id), "user_name")).value["data"]
#         return [SystemMessage(system_str)] + state["messages"]

#     async def prompt_no_store(state, config):
#         return SystemMessage("foo") + state["messages"]

#     model = FakeToolCallingModel()

#     # test state modifier that uses store works
#     agent = create_agent(model, [add], system_prompt=prompt, store=in_memory_store)
#     response = await agent.ainvoke(
#         {"messages": [("user", "hi")]}, {"configurable": {"user_id": "1"}}
#     )
#     assert response["messages"][-1].content == "User name is Alice-hi"

#     # test state modifier that doesn't use store works
#     agent = create_agent(model, [add], system_prompt=prompt_no_store, store=in_memory_store)
#     response = await agent.ainvoke(
#         {"messages": [("user", "hi")]}, {"configurable": {"user_id": "2"}}
#     )
#     assert response["messages"][-1].content == "foo-hi"


# @pytest.mark.parametrize("tool_style", ["openai", "anthropic"])
# @pytest.mark.parametrize("include_builtin", [True, False])
# def test_model_with_tools(tool_style: str, include_builtin: bool) -> None:
#     model = FakeToolCallingModel(tool_style=tool_style)

#     @dec_tool
#     def tool1(some_val: int) -> str:
#         """Tool 1 docstring."""
#         return f"Tool 1: {some_val}"

#     @dec_tool
#     def tool2(some_val: int) -> str:
#         """Tool 2 docstring."""
#         return f"Tool 2: {some_val}"

#     tools: list[BaseTool | dict] = [tool1, tool2]
#     if include_builtin:
#         tools.append(
#             {
#                 "type": "mcp",
#                 "server_label": "atest_sever",
#                 "server_url": "https://some.mcp.somewhere.com/sse",
#                 "headers": {"foo": "bar"},
#                 "allowed_tools": [
#                     "mcp_tool_1",
#                     "set_active_account",
#                     "get_url_markdown",
#                     "get_url_screenshot",
#                 ],
#                 "require_approval": "never",
#             }
#         )
#     # check valid agent constructor
#     with pytest.raises(ValueError):
#         create_agent(
#             model.bind_tools(tools),
#             tools,
#         )


# # Test removed: _validate_chat_history function no longer exists
# # def test__validate_messages() -> None:
# #     pass


# def test__infer_handled_types() -> None:
#     def handle(e) -> str:  # type: ignore
#         return ""

#     def handle2(e: Exception) -> str:
#         return ""

#     def handle3(e: ValueError | ToolException) -> str:
#         return ""

#     def handle4(e: Union[ValueError, ToolException]) -> str:
#         return ""

#     class Handler:
#         def handle(self, e: ValueError) -> str:
#             return ""

#     handle5 = Handler().handle

#     def handle6(e: Union[Union[TypeError, ValueError], ToolException]) -> str:
#         return ""

#     expected: tuple = (Exception,)
#     actual = _infer_handled_types(handle)
#     assert expected == actual

#     expected = (Exception,)
#     actual = _infer_handled_types(handle2)
#     assert expected == actual

#     expected = (ValueError, ToolException)
#     actual = _infer_handled_types(handle3)
#     assert expected == actual

#     expected = (ValueError, ToolException)
#     actual = _infer_handled_types(handle4)
#     assert expected == actual

#     expected = (ValueError,)
#     actual = _infer_handled_types(handle5)
#     assert expected == actual

#     expected = (TypeError, ValueError, ToolException)
#     actual = _infer_handled_types(handle6)
#     assert expected == actual

#     with pytest.raises(ValueError):

#         def handler(e: str) -> str:
#             return ""

#         _infer_handled_types(handler)

#     with pytest.raises(ValueError):

#         def handler(e: list[Exception]) -> str:
#             return ""

#         _infer_handled_types(handler)

#     with pytest.raises(ValueError):

#         def handler(e: Union[str, int]) -> str:
#             return ""

#         _infer_handled_types(handler)


# def test_react_agent_with_structured_response() -> None:
#     class WeatherResponse(BaseModel):
#         temperature: float = Field(description="The temperature in fahrenheit")

#     tool_calls = [
#         [{"args": {}, "id": "1", "name": "get_weather"}],
#         [{"name": "WeatherResponse", "id": "2", "args": {"temperature": 75}}],
#     ]

#     def get_weather() -> str:
#         """Get the weather"""
#         return "The weather is sunny and 75°F."

#     expected_structured_response = WeatherResponse(temperature=75)
#     model = FakeToolCallingModel[WeatherResponse](
#         tool_calls=tool_calls, structured_response=expected_structured_response
#     )
#     agent = create_agent(
#         model,
#         [get_weather],
#         response_format=WeatherResponse,
#     )
#     response = agent.invoke({"messages": [HumanMessage("What's the weather?")]})
#     assert response["structured_response"] == expected_structured_response
#     assert len(response["messages"]) == 5

#     # Check message types in message history
#     msg_types = [m.type for m in response["messages"]]
#     assert msg_types == [
#         "human",  # "What's the weather?"
#         "ai",  # "What's the weather?"
#         "tool",  # "The weather is sunny and 75°F."
#         "ai",  # structured response
#         "tool",  # artificial tool message
#     ]

#     assert [m.content for m in response["messages"]] == [
#         "What's the weather?",
#         "What's the weather?",
#         "The weather is sunny and 75°F.",
#         "What's the weather?-What's the weather?-The weather is sunny and 75°F.",
#         "Returning structured response: {'temperature': 75.0}",
#     ]


# class CustomState(AgentState):
#     user_name: str


# def test_react_agent_update_state(
#     sync_checkpointer: BaseCheckpointSaver,
# ) -> None:
#     @dec_tool
#     def get_user_name(tool_call_id: Annotated[str, InjectedToolCallId]):
#         """Retrieve user name"""
#         user_name = interrupt("Please provider user name:")
#         return Command(
#             update={
#                 "user_name": user_name,
#                 "messages": [
#                     ToolMessage("Successfully retrieved user name", tool_call_id=tool_call_id)
#                 ],
#             }
#         )

#     def prompt(state: CustomState):
#         user_name = state.get("user_name")
#         if user_name is None:
#             return state["messages"]

#         system_msg = f"User name is {user_name}"
#         return [{"role": "system", "content": system_msg}] + state["messages"]

#     tool_calls = [[{"args": {}, "id": "1", "name": "get_user_name"}]]
#     model = FakeToolCallingModel(tool_calls=tool_calls)
#     agent = create_agent(
#         model,
#         [get_user_name],
#         state_schema=CustomState,
#         prompt=prompt,
#         checkpointer=sync_checkpointer,
#     )
#     config = {"configurable": {"thread_id": "1"}}
#     # Run until interrupted
#     agent.invoke({"messages": [("user", "what's my name")]}, config)
#     # supply the value for the interrupt
#     response = agent.invoke(Command(resume="Archibald"), config)
#     # confirm that the state was updated
#     assert response["user_name"] == "Archibald"
#     assert len(response["messages"]) == 4
#     tool_message: ToolMessage = response["messages"][-2]
#     assert tool_message.content == "Successfully retrieved user name"
#     assert tool_message.tool_call_id == "1"
#     assert tool_message.name == "get_user_name"


# def test_react_agent_parallel_tool_calls(
#     sync_checkpointer: BaseCheckpointSaver,
# ) -> None:
#     human_assistance_execution_count = 0

#     @dec_tool
#     def human_assistance(query: str) -> str:
#         """Request assistance from a human."""
#         nonlocal human_assistance_execution_count
#         human_response = interrupt({"query": query})
#         human_assistance_execution_count += 1
#         return human_response["data"]

#     get_weather_execution_count = 0

#     @dec_tool
#     def get_weather(location: str) -> str:
#         """Use this tool to get the weather."""
#         nonlocal get_weather_execution_count
#         get_weather_execution_count += 1
#         return "It's sunny!"

#     tool_calls = [
#         [
#             {"args": {"location": "sf"}, "id": "1", "name": "get_weather"},
#             {"args": {"query": "request help"}, "id": "2", "name": "human_assistance"},
#         ],
#         [],
#     ]
#     model = FakeToolCallingModel(tool_calls=tool_calls)
#     agent = create_agent(
#         model,
#         [human_assistance, get_weather],
#         checkpointer=sync_checkpointer,
#     )
#     config = {"configurable": {"thread_id": "1"}}
#     query = "Get user assistance and also check the weather"
#     message_types = []
#     for event in agent.stream({"messages": [("user", query)]}, config, stream_mode="values"):
#         if messages := event.get("messages"):
#             message_types.append([m.type for m in messages])

#     assert message_types == [
#         ["human"],
#         ["human", "ai"],
#         ["human", "ai", "tool"],
#     ]

#     # Resume
#     message_types = []
#     for event in agent.stream(Command(resume={"data": "Hello"}), config, stream_mode="values"):
#         if messages := event.get("messages"):
#             message_types.append([m.type for m in messages])

#     assert message_types == [
#         ["human", "ai"],
#         ["human", "ai", "tool", "tool"],
#         ["human", "ai", "tool", "tool", "ai"],
#     ]

#     assert human_assistance_execution_count == 1
#     assert get_weather_execution_count == 1


# class AgentStateExtraKey(AgentState):
#     foo: int


# def test_create_agent_inject_vars() -> None:
#     """Test that the agent can inject state and store into tool functions."""
#     store = InMemoryStore()
#     namespace = ("test",)
#     store.put(namespace, "test_key", {"bar": 3})

#     def tool1(
#         some_val: int,
#         state: Annotated[dict, InjectedState],
#         store: Annotated[BaseStore, InjectedStore()],
#     ) -> str:
#         """Tool 1 docstring."""
#         store_val = store.get(namespace, "test_key").value["bar"]
#         return some_val + state["foo"] + store_val

#     tool_call = {
#         "name": "tool1",
#         "args": {"some_val": 1},
#         "id": "some 0",
#         "type": "tool_call",
#     }
#     model = FakeToolCallingModel(tool_calls=[[tool_call], []])
#     agent = create_agent(
#         model,
#         ToolNode([tool1], handle_tool_errors=False),
#         state_schema=AgentStateExtraKey,
#         store=store,
#     )
#     result = agent.invoke({"messages": [{"role": "user", "content": "hi"}], "foo": 2})
#     assert result["messages"] == [
#         _AnyIdHumanMessage(content="hi"),
#         AIMessage(content="hi", tool_calls=[tool_call], id="0"),
#         _AnyIdToolMessage(content="6", name="tool1", tool_call_id="some 0"),
#         AIMessage("hi-hi-6", id="1"),
#     ]
#     assert result["foo"] == 2


# async def test_return_direct() -> None:
#     @dec_tool(return_direct=True)
#     def tool_return_direct(input: str) -> str:
#         """A tool that returns directly."""
#         return f"Direct result: {input}"

#     @dec_tool
#     def tool_normal(input: str) -> str:
#         """A normal tool."""
#         return f"Normal result: {input}"

#     first_tool_call = [
#         ToolCall(
#             name="tool_return_direct",
#             args={"input": "Test direct"},
#             id="1",
#         ),
#     ]
#     expected_ai = AIMessage(
#         content="Test direct",
#         id="0",
#         tool_calls=first_tool_call,
#     )
#     model = FakeToolCallingModel(tool_calls=[first_tool_call, []])
#     agent = create_agent(
#         model,
#         [tool_return_direct, tool_normal],
#     )

#     # Test direct return for tool_return_direct
#     result = agent.invoke({"messages": [HumanMessage(content="Test direct", id="hum0")]})
#     assert result["messages"] == [
#         HumanMessage(content="Test direct", id="hum0"),
#         expected_ai,
#         ToolMessage(
#             content="Direct result: Test direct",
#             name="tool_return_direct",
#             tool_call_id="1",
#             id=result["messages"][2].id,
#         ),
#     ]
#     second_tool_call = [
#         ToolCall(
#             name="tool_normal",
#             args={"input": "Test normal"},
#             id="2",
#         ),
#     ]
#     model = FakeToolCallingModel(tool_calls=[second_tool_call, []])
#     agent = create_agent(model, [tool_return_direct, tool_normal])
#     result = agent.invoke({"messages": [HumanMessage(content="Test normal", id="hum1")]})
#     assert result["messages"] == [
#         HumanMessage(content="Test normal", id="hum1"),
#         AIMessage(content="Test normal", id="0", tool_calls=second_tool_call),
#         ToolMessage(
#             content="Normal result: Test normal",
#             name="tool_normal",
#             tool_call_id="2",
#             id=result["messages"][2].id,
#         ),
#         AIMessage(content="Test normal-Test normal-Normal result: Test normal", id="1"),
#     ]

#     both_tool_calls = [
#         ToolCall(
#             name="tool_return_direct",
#             args={"input": "Test both direct"},
#             id="3",
#         ),
#         ToolCall(
#             name="tool_normal",
#             args={"input": "Test both normal"},
#             id="4",
#         ),
#     ]
#     model = FakeToolCallingModel(tool_calls=[both_tool_calls, []])
#     agent = create_agent(model, [tool_return_direct, tool_normal])
#     result = agent.invoke({"messages": [HumanMessage(content="Test both", id="hum2")]})
#     assert result["messages"] == [
#         HumanMessage(content="Test both", id="hum2"),
#         AIMessage(content="Test both", id="0", tool_calls=both_tool_calls),
#         ToolMessage(
#             content="Direct result: Test both direct",
#             name="tool_return_direct",
#             tool_call_id="3",
#             id=result["messages"][2].id,
#         ),
#         ToolMessage(
#             content="Normal result: Test both normal",
#             name="tool_normal",
#             tool_call_id="4",
#             id=result["messages"][3].id,
#         ),
#     ]


# def test__get_state_args() -> None:
#     class Schema1(BaseModel):
#         a: Annotated[str, InjectedState]

#     class Schema2(Schema1):
#         b: Annotated[int, InjectedState("bar")]

#     @dec_tool(args_schema=Schema2)
#     def foo(a: str, b: int) -> float:
#         """return"""
#         return 0.0

#     assert _get_state_args(foo) == {"a": None, "b": "bar"}


# def test_inspect_react() -> None:
#     model = FakeToolCallingModel(tool_calls=[])
#     agent = create_agent(model, [])
#     inspect.getclosurevars(agent.nodes["agent"].bound.func)


# def test_react_with_subgraph_tools(
#     sync_checkpointer: BaseCheckpointSaver,
# ) -> None:
#     class State(TypedDict):
#         a: int
#         b: int

#     class Output(TypedDict):
#         result: int

#     # Define the subgraphs
#     def add(state):
#         return {"result": state["a"] + state["b"]}

#     add_subgraph = (
#         StateGraph(State, output_schema=Output).add_node(add).add_edge(START, "add").compile()
#     )

#     def multiply(state):
#         return {"result": state["a"] * state["b"]}

#     multiply_subgraph = (
#         StateGraph(State, output_schema=Output)
#         .add_node(multiply)
#         .add_edge(START, "multiply")
#         .compile()
#     )

#     multiply_subgraph.invoke({"a": 2, "b": 3})

#     # Add subgraphs as tools

#     def addition(a: int, b: int):
#         """Add two numbers"""
#         return add_subgraph.invoke({"a": a, "b": b})["result"]

#     def multiplication(a: int, b: int):
#         """Multiply two numbers"""
#         return multiply_subgraph.invoke({"a": a, "b": b})["result"]

#     model = FakeToolCallingModel(
#         tool_calls=[
#             [
#                 {"args": {"a": 2, "b": 3}, "id": "1", "name": "addition"},
#                 {"args": {"a": 2, "b": 3}, "id": "2", "name": "multiplication"},
#             ],
#             [],
#         ]
#     )
#     tool_node = ToolNode([addition, multiplication], handle_tool_errors=False)
#     agent = create_agent(
#         model,
#         tool_node,
#         checkpointer=sync_checkpointer,
#     )
#     result = agent.invoke(
#         {"messages": [HumanMessage(content="What's 2 + 3 and 2 * 3?")]},
#         config={"configurable": {"thread_id": "1"}},
#     )
#     assert result["messages"] == [
#         _AnyIdHumanMessage(content="What's 2 + 3 and 2 * 3?"),
#         AIMessage(
#             content="What's 2 + 3 and 2 * 3?",
#             id="0",
#             tool_calls=[
#                 ToolCall(name="addition", args={"a": 2, "b": 3}, id="1"),
#                 ToolCall(name="multiplication", args={"a": 2, "b": 3}, id="2"),
#             ],
#         ),
#         ToolMessage(content="5", name="addition", tool_call_id="1", id=result["messages"][2].id),
#         ToolMessage(
#             content="6",
#             name="multiplication",
#             tool_call_id="2",
#             id=result["messages"][3].id,
#         ),
#         AIMessage(content="What's 2 + 3 and 2 * 3?-What's 2 + 3 and 2 * 3?-5-6", id="1"),
#     ]


# def test_react_agent_subgraph_streaming_sync() -> None:
#     """Test React agent streaming when used as a subgraph node sync version"""

#     @dec_tool
#     def get_weather(city: str) -> str:
#         """Get the weather of a city."""
#         return f"The weather of {city} is sunny."

#     # Create a React agent
#     model = FakeToolCallingModel(
#         tool_calls=[
#             [{"args": {"city": "Tokyo"}, "id": "1", "name": "get_weather"}],
#             [],
#         ]
#     )

#     agent = create_agent(
#         model,
#         tools=[get_weather],
#         prompt="You are a helpful travel assistant.",
#     )

#     # Create a subgraph that uses the React agent as a node
#     def react_agent_node(state: MessagesState, config: RunnableConfig) -> MessagesState:
#         """Node that runs the React agent and collects streaming output."""
#         collected_content = ""

#         # Stream the agent output and collect content
#         for msg_chunk, _msg_metadata in agent.stream(
#             {"messages": [("user", state["messages"][-1].content)]},
#             config,
#             stream_mode="messages",
#         ):
#             if hasattr(msg_chunk, "content") and msg_chunk.content:
#                 collected_content += msg_chunk.content

#         return {"messages": [("assistant", collected_content)]}

#     # Create the main workflow with the React agent as a subgraph node
#     workflow = StateGraph(MessagesState)
#     workflow.add_node("react_agent", react_agent_node)
#     workflow.add_edge(START, "react_agent")
#     workflow.add_edge("react_agent", "__end__")
#     compiled_workflow = workflow.compile()

#     # Test the streaming functionality
#     result = compiled_workflow.invoke({"messages": [("user", "What is the weather in Tokyo?")]})

#     # Verify the result contains expected structure
#     assert len(result["messages"]) == 2
#     assert result["messages"][0].content == "What is the weather in Tokyo?"
#     assert "assistant" in str(result["messages"][1])

#     # Test streaming with subgraphs = True
#     result = compiled_workflow.invoke(
#         {"messages": [("user", "What is the weather in Tokyo?")]},
#         subgraphs=True,
#     )
#     assert len(result["messages"]) == 2

#     events = []
#     for event in compiled_workflow.stream(
#         {"messages": [("user", "What is the weather in Tokyo?")]},
#         stream_mode="messages",
#         subgraphs=False,
#     ):
#         events.append(event)

#     assert len(events) == 0

#     events = []
#     for event in compiled_workflow.stream(
#         {"messages": [("user", "What is the weather in Tokyo?")]},
#         stream_mode="messages",
#         subgraphs=True,
#     ):
#         events.append(event)

#     assert len(events) == 3
#     namespace, (msg, metadata) = events[0]
#     # FakeToolCallingModel returns a single AIMessage with tool calls
#     # The content of the AIMessage reflects the input message
#     assert msg.content.startswith("You are a helpful travel assistant")
#     namespace, (msg, metadata) = events[1]  # ToolMessage
#     assert msg.content.startswith("The weather of Tokyo is sunny.")


# async def test_react_agent_subgraph_streaming() -> None:
#     """Test React agent streaming when used as a subgraph node."""

#     @dec_tool
#     def get_weather(city: str) -> str:
#         """Get the weather of a city."""
#         return f"The weather of {city} is sunny."

#     # Create a React agent
#     model = FakeToolCallingModel(
#         tool_calls=[
#             [{"args": {"city": "Tokyo"}, "id": "1", "name": "get_weather"}],
#             [],
#         ]
#     )

#     agent = create_agent(
#         model,
#         tools=[get_weather],
#         prompt="You are a helpful travel assistant.",
#     )

#     # Create a subgraph that uses the React agent as a node
#     async def react_agent_node(state: MessagesState, config: RunnableConfig) -> MessagesState:
#         """Node that runs the React agent and collects streaming output."""
#         collected_content = ""

#         # Stream the agent output and collect content
#         async for msg_chunk, _msg_metadata in agent.astream(
#             {"messages": [("user", state["messages"][-1].content)]},
#             config,
#             stream_mode="messages",
#         ):
#             if hasattr(msg_chunk, "content") and msg_chunk.content:
#                 collected_content += msg_chunk.content

#         return {"messages": [("assistant", collected_content)]}

#     # Create the main workflow with the React agent as a subgraph node
#     workflow = StateGraph(MessagesState)
#     workflow.add_node("react_agent", react_agent_node)
#     workflow.add_edge(START, "react_agent")
#     workflow.add_edge("react_agent", "__end__")
#     compiled_workflow = workflow.compile()

#     # Test the streaming functionality
#     result = await compiled_workflow.ainvoke(
#         {"messages": [("user", "What is the weather in Tokyo?")]}
#     )

#     # Verify the result contains expected structure
#     assert len(result["messages"]) == 2
#     assert result["messages"][0].content == "What is the weather in Tokyo?"
#     assert "assistant" in str(result["messages"][1])

#     # Test streaming with subgraphs = True
#     result = await compiled_workflow.ainvoke(
#         {"messages": [("user", "What is the weather in Tokyo?")]},
#         subgraphs=True,
#     )
#     assert len(result["messages"]) == 2

#     events = []
#     async for event in compiled_workflow.astream(
#         {"messages": [("user", "What is the weather in Tokyo?")]},
#         stream_mode="messages",
#         subgraphs=False,
#     ):
#         events.append(event)

#     assert len(events) == 0

#     events = []
#     async for event in compiled_workflow.astream(
#         {"messages": [("user", "What is the weather in Tokyo?")]},
#         stream_mode="messages",
#         subgraphs=True,
#     ):
#         events.append(event)

#     assert len(events) == 3
#     namespace, (msg, metadata) = events[0]
#     # FakeToolCallingModel returns a single AIMessage with tool calls
#     # The content of the AIMessage reflects the input message
#     assert msg.content.startswith("You are a helpful travel assistant")
#     namespace, (msg, metadata) = events[1]  # ToolMessage
#     assert msg.content.startswith("The weather of Tokyo is sunny.")


# def test_tool_node_node_interrupt(
#     sync_checkpointer: BaseCheckpointSaver,
# ) -> None:
#     def tool_normal(some_val: int) -> str:
#         """Tool docstring."""
#         return "normal"

#     def tool_interrupt(some_val: int) -> str:
#         """Tool docstring."""
#         return interrupt("provide value for foo")

#     # test inside react agent
#     model = FakeToolCallingModel(
#         tool_calls=[
#             [
#                 ToolCall(name="tool_interrupt", args={"some_val": 0}, id="1"),
#                 ToolCall(name="tool_normal", args={"some_val": 1}, id="2"),
#             ],
#             [],
#         ]
#     )
#     config = {"configurable": {"thread_id": "1"}}
#     agent = create_agent(
#         model,
#         [tool_interrupt, tool_normal],
#         checkpointer=sync_checkpointer,
#     )
#     result = agent.invoke({"messages": [HumanMessage("hi?")]}, config)
#     expected_messages = [
#         _AnyIdHumanMessage(content="hi?"),
#         AIMessage(
#             content="hi?",
#             id="0",
#             tool_calls=[
#                 {
#                     "name": "tool_interrupt",
#                     "args": {"some_val": 0},
#                     "id": "1",
#                     "type": "tool_call",
#                 },
#                 {
#                     "name": "tool_normal",
#                     "args": {"some_val": 1},
#                     "id": "2",
#                     "type": "tool_call",
#                 },
#             ],
#         ),
#         _AnyIdToolMessage(content="normal", name="tool_normal", tool_call_id="2"),
#     ]
#     assert result["messages"] == expected_messages

#     state = agent.get_state(config)
#     assert state.next == ("tools",)
#     task = state.tasks[0]
#     assert task.name == "tools"
#     assert task.interrupts == (
#         Interrupt(
#             value="provide value for foo",
#             id=AnyStr(),
#         ),
#     )



"""Unit tests for ToolCallLimitMiddleware."""

import pytest
from langchain_core.messages import AIMessage, HumanMessage, ToolCall, ToolMessage
from langchain_core.tools import tool
from langgraph.checkpoint.memory import InMemorySaver

from langchain.agents.factory import create_agent
from langchain.agents.middleware.tool_call_limit import (
    ToolCallLimitExceededError,
    ToolCallLimitMiddleware,
    _count_tool_calls_in_messages,
    _get_run_messages,
)
from tests.unit_tests.agents.test_middleware_agent import FakeToolCallingModel


def test_count_tool_calls_in_messages():
    """Test counting tool calls in messages."""
    # Test with no messages
    assert _count_tool_calls_in_messages([]) == 0

    # Test with messages but no tool calls
    messages = [
        HumanMessage("Hello"),
        AIMessage("Hi there"),
    ]
    assert _count_tool_calls_in_messages(messages) == 0

    # Test with tool calls
    messages = [
        HumanMessage("Search for something"),
        AIMessage(
            "Searching...",
            tool_calls=[
                {"name": "search", "args": {"query": "test"}, "id": "1"},
                {"name": "calculator", "args": {"expression": "1+1"}, "id": "2"},
            ],
        ),
        ToolMessage("Result 1", tool_call_id="1"),
        ToolMessage("Result 2", tool_call_id="2"),
        AIMessage(
            "More searching...",
            tool_calls=[
                {"name": "search", "args": {"query": "another"}, "id": "3"},
            ],
        ),
    ]
    # Total: 3 tool calls (2 from first AI message, 1 from second)
    assert _count_tool_calls_in_messages(messages) == 3

    # Test filtering by tool name
    assert _count_tool_calls_in_messages(messages, tool_name="search") == 2
    assert _count_tool_calls_in_messages(messages, tool_name="calculator") == 1
    assert _count_tool_calls_in_messages(messages, tool_name="nonexistent") == 0


def test_get_run_messages():
    """Test extracting run messages after last HumanMessage."""
    # Test with no messages
    assert _get_run_messages([]) == []

    # Test with no HumanMessage
    messages = [
        AIMessage("Hello"),
        AIMessage("World"),
    ]
    assert _get_run_messages(messages) == messages

    # Test with HumanMessage at the end
    messages = [
        AIMessage("Previous"),
        HumanMessage("New question"),
    ]
    assert _get_run_messages(messages) == []

    # Test with messages after HumanMessage
    messages = [
        AIMessage("Previous"),
        ToolMessage("Previous result", tool_call_id="0"),
        HumanMessage("New question"),
        AIMessage(
            "Response",
            tool_calls=[{"name": "search", "args": {}, "id": "1"}],
        ),
        ToolMessage("Result", tool_call_id="1"),
    ]
    run_messages = _get_run_messages(messages)
    assert len(run_messages) == 2
    assert isinstance(run_messages[0], AIMessage)
    assert isinstance(run_messages[1], ToolMessage)


def test_middleware_initialization_validation():
    """Test that middleware initialization validates parameters correctly."""
    # Test that at least one limit must be specified
    with pytest.raises(ValueError, match="At least one limit must be specified"):
        ToolCallLimitMiddleware()

    # Test invalid exit behavior
    with pytest.raises(ValueError, match="Invalid exit_behavior"):
        ToolCallLimitMiddleware(thread_limit=5, exit_behavior="invalid")  # type: ignore[arg-type]

    # Test valid initialization
    middleware = ToolCallLimitMiddleware(thread_limit=5, run_limit=3)
    assert middleware.thread_limit == 5
    assert middleware.run_limit == 3
    assert middleware.exit_behavior == "end"
    assert middleware.tool_name is None

    # Test with tool name
    middleware = ToolCallLimitMiddleware(tool_name="search", thread_limit=5)
    assert middleware.tool_name == "search"
    assert middleware.thread_limit == 5
    assert middleware.run_limit is None


def test_middleware_name_property():
    """Test that the name property includes tool name when specified."""
    # Test without tool name
    middleware = ToolCallLimitMiddleware(thread_limit=5)
    assert middleware.name == "ToolCallLimitMiddleware"

    # Test with tool name
    middleware = ToolCallLimitMiddleware(tool_name="search", thread_limit=5)
    assert middleware.name == "ToolCallLimitMiddleware[search]"

    # Test multiple instances with different tool names have unique names
    middleware1 = ToolCallLimitMiddleware(tool_name="search", thread_limit=5)
    middleware2 = ToolCallLimitMiddleware(tool_name="calculator", thread_limit=3)
    assert middleware1.name != middleware2.name
    assert middleware1.name == "ToolCallLimitMiddleware[search]"
    assert middleware2.name == "ToolCallLimitMiddleware[calculator]"


def test_middleware_unit_functionality():
    """Test that the middleware works as expected in isolation."""
    # Test with end behavior - global tool limit
    middleware = ToolCallLimitMiddleware(thread_limit=3, run_limit=2)

    # Mock runtime (not used in current implementation)
    runtime = None

    # Test when limits are not exceeded
    state = {
        "messages": [
            HumanMessage("Question"),
            AIMessage(
                "Response",
                tool_calls=[{"name": "search", "args": {}, "id": "1"}],
            ),
        ]
    }
    result = middleware.before_model(state, runtime)  # type: ignore[arg-type]
    assert result is None

    # Test when thread limit is exceeded
    state = {
        "messages": [
            HumanMessage("Question 1"),
            AIMessage(
                "Response 1",
                tool_calls=[
                    {"name": "search", "args": {}, "id": "1"},
                    {"name": "calculator", "args": {}, "id": "2"},
                ],
            ),
            ToolMessage("Result 1", tool_call_id="1"),
            ToolMessage("Result 2", tool_call_id="2"),
            HumanMessage("Question 2"),
            AIMessage(
                "Response 2",
                tool_calls=[{"name": "search", "args": {}, "id": "3"}],
            ),
        ]
    }
    result = middleware.before_model(state, runtime)  # type: ignore[arg-type]
    assert result is not None
    assert result["jump_to"] == "end"
    assert "messages" in result
    assert len(result["messages"]) == 1
    assert "thread limit (3/3)" in result["messages"][0].content

    # Test when run limit is exceeded
    state = {
        "messages": [
            HumanMessage("Question"),
            AIMessage(
                "Response",
                tool_calls=[
                    {"name": "search", "args": {}, "id": "1"},
                    {"name": "calculator", "args": {}, "id": "2"},
                ],
            ),
        ]
    }
    result = middleware.before_model(state, runtime)  # type: ignore[arg-type]
    assert result is not None
    assert result["jump_to"] == "end"
    assert "run limit (2/2)" in result["messages"][0].content


def test_middleware_with_specific_tool():
    """Test middleware that limits a specific tool."""
    # Limit only the "search" tool
    middleware = ToolCallLimitMiddleware(tool_name="search", thread_limit=2, run_limit=1)

    runtime = None

    # Test with search tool calls
    state = {
        "messages": [
            HumanMessage("Question"),
            AIMessage(
                "Response",
                tool_calls=[
                    {"name": "search", "args": {}, "id": "1"},
                    {"name": "calculator", "args": {}, "id": "2"},
                ],
            ),
        ]
    }
    # Run limit for search is 1, should be exceeded
    result = middleware.before_model(state, runtime)  # type: ignore[arg-type]
    assert result is not None
    assert "'search' tool call" in result["messages"][0].content
    assert "run limit (1/1)" in result["messages"][0].content

    # Test with only calculator calls (should not trigger limit)
    state = {
        "messages": [
            HumanMessage("Question"),
            AIMessage(
                "Response",
                tool_calls=[
                    {"name": "calculator", "args": {}, "id": "1"},
                    {"name": "calculator", "args": {}, "id": "2"},
                ],
            ),
        ]
    }
    result = middleware.before_model(state, runtime)  # type: ignore[arg-type]
    assert result is None


def test_middleware_error_behavior():
    """Test middleware with error exit behavior."""
    middleware = ToolCallLimitMiddleware(thread_limit=2, run_limit=1, exit_behavior="error")

    runtime = None

    # Test exception when thread limit exceeded
    state = {
        "messages": [
            HumanMessage("Q"),
            AIMessage("R", tool_calls=[{"name": "search", "args": {}, "id": "1"}]),
            ToolMessage("Result", tool_call_id="1"),
            AIMessage("R", tool_calls=[{"name": "search", "args": {}, "id": "2"}]),
        ]
    }
    with pytest.raises(ToolCallLimitExceededError) as exc_info:
        middleware.before_model(state, runtime)  # type: ignore[arg-type]

    assert "thread limit (2/2)" in str(exc_info.value)
    assert exc_info.value.thread_count == 2
    assert exc_info.value.thread_limit == 2

    # Test exception when run limit exceeded
    state = {
        "messages": [
            HumanMessage("Question"),
            AIMessage("R", tool_calls=[{"name": "search", "args": {}, "id": "1"}]),
        ]
    }
    with pytest.raises(ToolCallLimitExceededError) as exc_info:
        middleware.before_model(state, runtime)  # type: ignore[arg-type]

    assert "run limit (1/1)" in str(exc_info.value)


def test_multiple_middleware_instances():
    """Test using multiple ToolCallLimitMiddleware instances in the same agent."""

    @tool
    def search(query: str) -> str:
        """Search for information."""
        return f"Results for: {query}"

    @tool
    def calculator(expression: str) -> str:
        """Calculate an expression."""
        return f"Result: {expression}"

    # Create model that makes multiple tool calls
    model = FakeToolCallingModel(
        tool_calls=[
            [
                ToolCall(name="search", args={"query": "test1"}, id="1"),
                ToolCall(name="calculator", args={"expression": "1+1"}, id="2"),
            ],
            [
                ToolCall(name="search", args={"query": "test2"}, id="3"),
                ToolCall(name="calculator", args={"expression": "2+2"}, id="4"),
            ],
            [
                ToolCall(name="search", args={"query": "test3"}, id="5"),
            ],
            [],
        ]
    )

    # Global limit: max 5 tool calls per thread
    # Search-specific limit: max 2 search calls per thread
    global_limiter = ToolCallLimitMiddleware(thread_limit=5)
    search_limiter = ToolCallLimitMiddleware(tool_name="search", thread_limit=2)

    agent = create_agent(
        model=model,
        tools=[search, calculator],
        middleware=[global_limiter, search_limiter],
        checkpointer=InMemorySaver(),
    )

    # First invocation: 2 search calls, 2 calculator calls (4 total)
    # Should succeed - within both limits
    result = agent.invoke(
        {"messages": [HumanMessage("First question")]},
        {"configurable": {"thread_id": "test_thread"}},
    )
    assert len(result["messages"]) > 0

    # Second invocation would add 1 more search call (total 3), exceeding search limit
    # Should hit the search-specific limit
    result = agent.invoke(
        {"messages": [HumanMessage("Second question")]},
        {"configurable": {"thread_id": "test_thread"}},
    )

    # Check that the limit was hit
    last_message = result["messages"][-1]
    assert isinstance(last_message, AIMessage)
    assert "'search' tool call" in last_message.content
    assert "thread limit" in last_message.content


def test_run_limit_with_multiple_human_messages():
    """Test that run limit correctly resets after each HumanMessage."""

    @tool
    def search(query: str) -> str:
        """Search for information."""
        return f"Results for: {query}"

    # Model makes 2 tool calls, then 1 tool call, then no tool calls
    model = FakeToolCallingModel(
        tool_calls=[
            [
                ToolCall(name="search", args={"query": "test1"}, id="1"),
                ToolCall(name="search", args={"query": "test2"}, id="2"),
            ],
            [ToolCall(name="search", args={"query": "test3"}, id="3")],
            [],
        ]
    )

    # Run limit of 2 tool calls per run
    middleware = ToolCallLimitMiddleware(run_limit=2)

    agent = create_agent(
        model=model, tools=[search], middleware=[middleware], checkpointer=InMemorySaver()
    )

    thread_config = {"configurable": {"thread_id": "test_thread"}}

    # First run: 2 tool calls, should hit run limit
    result = agent.invoke({"messages": [HumanMessage("Question 1")]}, thread_config)
    last_message = result["messages"][-1]
    assert "run limit (2/2)" in last_message.content

    # Second run: starts fresh, should also hit run limit after 2 calls
    result = agent.invoke({"messages": [HumanMessage("Question 2")]}, thread_config)
    last_message = result["messages"][-1]
    # This should also hit the limit because run count resets
    assert "run limit (2/2)" in last_message.content


def test_exception_error_messages():
    """Test that exceptions provide clear error messages."""
    # Test global tool limit
    middleware = ToolCallLimitMiddleware(thread_limit=3, run_limit=2, exit_behavior="error")

    state = {
        "messages": [
            HumanMessage("Q"),
            AIMessage("R", tool_calls=[{"name": "search", "args": {}, "id": "1"}]),
            ToolMessage("Result", tool_call_id="1"),
            AIMessage("R", tool_calls=[{"name": "search", "args": {}, "id": "2"}]),
            ToolMessage("Result", tool_call_id="2"),
            AIMessage("R", tool_calls=[{"name": "search", "args": {}, "id": "3"}]),
        ]
    }

    with pytest.raises(ToolCallLimitExceededError) as exc_info:
        middleware.before_model(state, None)  # type: ignore[arg-type]

    error_msg = str(exc_info.value)
    assert "Tool call limits exceeded" in error_msg
    assert "thread limit (3/3)" in error_msg

    # Test specific tool limit
    middleware = ToolCallLimitMiddleware(tool_name="search", thread_limit=2, exit_behavior="error")

    state = {
        "messages": [
            HumanMessage("Q"),
            AIMessage(
                "R",
                tool_calls=[
                    {"name": "search", "args": {}, "id": "1"},
                    {"name": "calculator", "args": {}, "id": "2"},
                ],
            ),
            ToolMessage("Result", tool_call_id="1"),
            ToolMessage("Result", tool_call_id="2"),
            AIMessage("R", tool_calls=[{"name": "search", "args": {}, "id": "3"}]),
        ]
    }

    with pytest.raises(ToolCallLimitExceededError) as exc_info:
        middleware.before_model(state, None)  # type: ignore[arg-type]

    error_msg = str(exc_info.value)
    assert "'search' tool call limits exceeded" in error_msg
    assert "thread limit (2/2)" in error_msg



"""Consolidated tests for middleware decorators: before_model, after_model, and modify_model_request."""

import pytest
from typing import Any
from typing_extensions import NotRequired
from syrupy.assertion import SnapshotAssertion

from langchain_core.messages import HumanMessage, AIMessage
from langchain_core.tools import tool
from langgraph.runtime import Runtime
from langgraph.types import Command

from langchain.agents.middleware.types import (
    AgentMiddleware,
    AgentState,
    ModelRequest,
    before_model,
    after_model,
    dynamic_prompt,
    modify_model_request,
    hook_config,
)
from langchain.agents.factory import create_agent, _get_can_jump_to
from .model import FakeToolCallingModel


class CustomState(AgentState):
    """Custom state schema for testing."""

    custom_field: NotRequired[str]


@tool
def test_tool(input: str) -> str:
    """A test tool for middleware testing."""
    return f"Tool result: {input}"


def test_before_model_decorator() -> None:
    """Test before_model decorator with all configuration options."""

    @before_model(
        state_schema=CustomState, tools=[test_tool], can_jump_to=["end"], name="CustomBeforeModel"
    )
    def custom_before_model(state: CustomState, runtime: Runtime) -> dict[str, Any]:
        return {"jump_to": "end"}

    assert isinstance(custom_before_model, AgentMiddleware)
    assert custom_before_model.state_schema == CustomState
    assert custom_before_model.tools == [test_tool]
    assert getattr(custom_before_model.__class__.before_model, "__can_jump_to__", []) == ["end"]
    assert custom_before_model.__class__.__name__ == "CustomBeforeModel"

    result = custom_before_model.before_model({"messages": [HumanMessage("Hello")]}, None)
    assert result == {"jump_to": "end"}


def test_after_model_decorator() -> None:
    """Test after_model decorator with all configuration options."""

    @after_model(
        state_schema=CustomState,
        tools=[test_tool],
        can_jump_to=["model", "end"],
        name="CustomAfterModel",
    )
    def custom_after_model(state: CustomState, runtime: Runtime) -> dict[str, Any]:
        return {"jump_to": "model"}

    # Verify all options were applied
    assert isinstance(custom_after_model, AgentMiddleware)
    assert custom_after_model.state_schema == CustomState
    assert custom_after_model.tools == [test_tool]
    assert getattr(custom_after_model.__class__.after_model, "__can_jump_to__", []) == [
        "model",
        "end",
    ]
    assert custom_after_model.__class__.__name__ == "CustomAfterModel"

    # Verify it works
    result = custom_after_model.after_model(
        {"messages": [HumanMessage("Hello"), AIMessage("Hi!")]}, None
    )
    assert result == {"jump_to": "model"}


def test_modify_model_request_decorator() -> None:
    """Test modify_model_request decorator with all configuration options."""

    @modify_model_request(state_schema=CustomState, tools=[test_tool], name="CustomModifyRequest")
    def custom_modify_request(
        request: ModelRequest, state: CustomState, runtime: Runtime
    ) -> ModelRequest:
        request.system_prompt = "Modified"
        return request

    # Verify all options were applied
    assert isinstance(custom_modify_request, AgentMiddleware)
    assert custom_modify_request.state_schema == CustomState
    assert custom_modify_request.tools == [test_tool]
    assert custom_modify_request.__class__.__name__ == "CustomModifyRequest"

    # Verify it works
    original_request = ModelRequest(
        model="test-model",
        system_prompt="Original",
        messages=[HumanMessage("Hello")],
        tool_choice=None,
        tools=[],
        response_format=None,
    )
    result = custom_modify_request.modify_model_request(
        original_request, {"messages": [HumanMessage("Hello")]}, None
    )
    assert result.system_prompt == "Modified"


def test_all_decorators_integration() -> None:
    """Test all three decorators working together in an agent."""
    call_order = []

    @before_model
    def track_before(state: AgentState, runtime: Runtime) -> None:
        call_order.append("before")
        return None

    @modify_model_request
    def track_modify(request: ModelRequest, state: AgentState, runtime: Runtime) -> ModelRequest:
        call_order.append("modify")
        return request

    @after_model
    def track_after(state: AgentState, runtime: Runtime) -> None:
        call_order.append("after")
        return None

    agent = create_agent(
        model=FakeToolCallingModel(), middleware=[track_before, track_modify, track_after]
    )
    # Agent is already compiled
    agent.invoke({"messages": [HumanMessage("Hello")]})

    assert call_order == ["before", "modify", "after"]


def test_decorators_use_function_names_as_default() -> None:
    """Test that decorators use function names as default middleware names."""

    @before_model
    def my_before_hook(state: AgentState, runtime: Runtime) -> None:
        return None

    @modify_model_request
    def my_modify_hook(request: ModelRequest, state: AgentState, runtime: Runtime) -> ModelRequest:
        return request

    @after_model
    def my_after_hook(state: AgentState, runtime: Runtime) -> None:
        return None

    # Verify that function names are used as middleware class names
    assert my_before_hook.__class__.__name__ == "my_before_hook"
    assert my_modify_hook.__class__.__name__ == "my_modify_hook"
    assert my_after_hook.__class__.__name__ == "my_after_hook"


def test_hook_config_decorator_on_class_method() -> None:
    """Test hook_config decorator on AgentMiddleware class methods."""

    class JumpMiddleware(AgentMiddleware):
        @hook_config(can_jump_to=["end", "model"])
        def before_model(self, state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
            if len(state["messages"]) > 5:
                return {"jump_to": "end"}
            return None

        @hook_config(can_jump_to=["tools"])
        def after_model(self, state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
            return {"jump_to": "tools"}

    # Verify can_jump_to metadata is preserved
    assert getattr(JumpMiddleware.before_model, "__can_jump_to__", []) == ["end", "model"]
    assert getattr(JumpMiddleware.after_model, "__can_jump_to__", []) == ["tools"]


def test_can_jump_to_with_before_model_decorator() -> None:
    """Test can_jump_to parameter used with before_model decorator."""

    @before_model(can_jump_to=["end"])
    def conditional_before(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        if len(state["messages"]) > 3:
            return {"jump_to": "end"}
        return None

    # Verify middleware was created and has can_jump_to metadata
    assert isinstance(conditional_before, AgentMiddleware)
    assert getattr(conditional_before.__class__.before_model, "__can_jump_to__", []) == ["end"]


def test_can_jump_to_with_after_model_decorator() -> None:
    """Test can_jump_to parameter used with after_model decorator."""

    @after_model(can_jump_to=["model", "end"])
    def conditional_after(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        if state["messages"][-1].content == "retry":
            return {"jump_to": "model"}
        return None

    # Verify middleware was created and has can_jump_to metadata
    assert isinstance(conditional_after, AgentMiddleware)
    assert getattr(conditional_after.__class__.after_model, "__can_jump_to__", []) == [
        "model",
        "end",
    ]


def test_can_jump_to_integration() -> None:
    """Test can_jump_to parameter in a full agent."""
    calls = []

    @before_model(can_jump_to=["end"])
    def early_exit(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        calls.append("early_exit")
        if state["messages"][0].content == "exit":
            return {"jump_to": "end"}
        return None

    agent = create_agent(model=FakeToolCallingModel(), middleware=[early_exit])
    # Agent is already compiled

    # Test with early exit
    result = agent.invoke({"messages": [HumanMessage("exit")]})
    assert calls == ["early_exit"]
    assert len(result["messages"]) == 1

    # Test without early exit
    calls.clear()
    result = agent.invoke({"messages": [HumanMessage("hello")]})
    assert calls == ["early_exit"]
    assert len(result["messages"]) > 1


# Async Decorator Tests


def test_async_before_model_decorator() -> None:
    """Test before_model decorator with async function."""

    @before_model(state_schema=CustomState, tools=[test_tool], name="AsyncBeforeModel")
    async def async_before_model(state: CustomState, runtime: Runtime) -> dict[str, Any]:
        return {"custom_field": "async_value"}

    assert isinstance(async_before_model, AgentMiddleware)
    assert async_before_model.state_schema == CustomState
    assert async_before_model.tools == [test_tool]
    assert async_before_model.__class__.__name__ == "AsyncBeforeModel"


def test_async_after_model_decorator() -> None:
    """Test after_model decorator with async function."""

    @after_model(state_schema=CustomState, tools=[test_tool], name="AsyncAfterModel")
    async def async_after_model(state: CustomState, runtime: Runtime) -> dict[str, Any]:
        return {"custom_field": "async_value"}

    assert isinstance(async_after_model, AgentMiddleware)
    assert async_after_model.state_schema == CustomState
    assert async_after_model.tools == [test_tool]
    assert async_after_model.__class__.__name__ == "AsyncAfterModel"


def test_async_modify_model_request_decorator() -> None:
    """Test modify_model_request decorator with async function."""

    @modify_model_request(state_schema=CustomState, tools=[test_tool], name="AsyncModifyRequest")
    async def async_modify_request(
        request: ModelRequest, state: CustomState, runtime: Runtime
    ) -> ModelRequest:
        request.system_prompt = "Modified async"
        return request

    assert isinstance(async_modify_request, AgentMiddleware)
    assert async_modify_request.state_schema == CustomState
    assert async_modify_request.tools == [test_tool]
    assert async_modify_request.__class__.__name__ == "AsyncModifyRequest"


def test_mixed_sync_async_decorators() -> None:
    """Test decorators with both sync and async functions."""

    @before_model(name="MixedBeforeModel")
    def sync_before(state: AgentState, runtime: Runtime) -> None:
        return None

    @before_model(name="MixedBeforeModel")
    async def async_before(state: AgentState, runtime: Runtime) -> None:
        return None

    @modify_model_request(name="MixedModifyRequest")
    def sync_modify(request: ModelRequest, state: AgentState, runtime: Runtime) -> ModelRequest:
        return request

    @modify_model_request(name="MixedModifyRequest")
    async def async_modify(
        request: ModelRequest, state: AgentState, runtime: Runtime
    ) -> ModelRequest:
        return request

    # Both should create valid middleware instances
    assert isinstance(sync_before, AgentMiddleware)
    assert isinstance(async_before, AgentMiddleware)
    assert isinstance(sync_modify, AgentMiddleware)
    assert isinstance(async_modify, AgentMiddleware)


@pytest.mark.asyncio
async def test_async_decorators_integration() -> None:
    """Test async decorators working together in an agent."""
    call_order = []

    @before_model
    async def track_async_before(state: AgentState, runtime: Runtime) -> None:
        call_order.append("async_before")
        return None

    @modify_model_request
    async def track_async_modify(
        request: ModelRequest, state: AgentState, runtime: Runtime
    ) -> ModelRequest:
        call_order.append("async_modify")
        return request

    @after_model
    async def track_async_after(state: AgentState, runtime: Runtime) -> None:
        call_order.append("async_after")
        return None

    agent = create_agent(
        model=FakeToolCallingModel(),
        middleware=[track_async_before, track_async_modify, track_async_after],
    )
    # Agent is already compiled
    await agent.ainvoke({"messages": [HumanMessage("Hello")]})

    assert call_order == ["async_before", "async_modify", "async_after"]


@pytest.mark.asyncio
async def test_mixed_sync_async_decorators_integration() -> None:
    """Test mixed sync/async decorators working together in an agent."""
    call_order = []

    @before_model
    def track_sync_before(state: AgentState, runtime: Runtime) -> None:
        call_order.append("sync_before")
        return None

    @before_model
    async def track_async_before(state: AgentState, runtime: Runtime) -> None:
        call_order.append("async_before")
        return None

    @modify_model_request
    def track_sync_modify(
        request: ModelRequest, state: AgentState, runtime: Runtime
    ) -> ModelRequest:
        call_order.append("sync_modify")
        return request

    @modify_model_request
    async def track_async_modify(
        request: ModelRequest, state: AgentState, runtime: Runtime
    ) -> ModelRequest:
        call_order.append("async_modify")
        return request

    @after_model
    async def track_async_after(state: AgentState, runtime: Runtime) -> None:
        call_order.append("async_after")
        return None

    @after_model
    def track_sync_after(state: AgentState, runtime: Runtime) -> None:
        call_order.append("sync_after")
        return None

    agent = create_agent(
        model=FakeToolCallingModel(),
        middleware=[
            track_sync_before,
            track_async_before,
            track_sync_modify,
            track_async_modify,
            track_async_after,
            track_sync_after,
        ],
    )
    # Agent is already compiled
    await agent.ainvoke({"messages": [HumanMessage("Hello")]})

    assert call_order == [
        "sync_before",
        "async_before",
        "sync_modify",
        "async_modify",
        "sync_after",
        "async_after",
    ]


def test_async_before_model_preserves_can_jump_to() -> None:
    """Test that can_jump_to metadata is preserved for async before_model functions."""

    @before_model(can_jump_to=["end"])
    async def async_conditional_before(
        state: AgentState, runtime: Runtime
    ) -> dict[str, Any] | None:
        if len(state["messages"]) > 3:
            return {"jump_to": "end"}
        return None

    # Verify middleware was created and has can_jump_to metadata
    assert isinstance(async_conditional_before, AgentMiddleware)
    assert getattr(async_conditional_before.__class__.abefore_model, "__can_jump_to__", []) == [
        "end"
    ]


def test_async_after_model_preserves_can_jump_to() -> None:
    """Test that can_jump_to metadata is preserved for async after_model functions."""

    @after_model(can_jump_to=["model", "end"])
    async def async_conditional_after(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        if state["messages"][-1].content == "retry":
            return {"jump_to": "model"}
        return None

    # Verify middleware was created and has can_jump_to metadata
    assert isinstance(async_conditional_after, AgentMiddleware)
    assert getattr(async_conditional_after.__class__.aafter_model, "__can_jump_to__", []) == [
        "model",
        "end",
    ]


@pytest.mark.asyncio
async def test_async_can_jump_to_integration() -> None:
    """Test can_jump_to parameter in a full agent with async middleware."""
    calls = []

    @before_model(can_jump_to=["end"])
    async def async_early_exit(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        calls.append("async_early_exit")
        if state["messages"][0].content == "exit":
            return {"jump_to": "end"}
        return None

    agent = create_agent(model=FakeToolCallingModel(), middleware=[async_early_exit])
    # Agent is already compiled

    # Test with early exit
    result = await agent.ainvoke({"messages": [HumanMessage("exit")]})
    assert calls == ["async_early_exit"]
    assert len(result["messages"]) == 1

    # Test without early exit
    calls.clear()
    result = await agent.ainvoke({"messages": [HumanMessage("hello")]})
    assert calls == ["async_early_exit"]
    assert len(result["messages"]) > 1


def test_get_can_jump_to_no_false_positives() -> None:
    """Test that _get_can_jump_to doesn't return false positives for base class methods."""

    # Middleware with no overridden methods should return empty list
    class EmptyMiddleware(AgentMiddleware):
        pass

    empty_middleware = EmptyMiddleware()
    empty_middleware.tools = []

    # Should not return any jump destinations for base class methods
    assert _get_can_jump_to(empty_middleware, "before_model") == []
    assert _get_can_jump_to(empty_middleware, "after_model") == []


def test_get_can_jump_to_only_overridden_methods() -> None:
    """Test that _get_can_jump_to only checks overridden methods."""

    # Middleware with only sync method overridden
    class SyncOnlyMiddleware(AgentMiddleware):
        @hook_config(can_jump_to=["end"])
        def before_model(self, state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
            return None

    sync_middleware = SyncOnlyMiddleware()
    sync_middleware.tools = []

    # Should return can_jump_to from overridden sync method
    assert _get_can_jump_to(sync_middleware, "before_model") == ["end"]

    # Middleware with only async method overridden
    class AsyncOnlyMiddleware(AgentMiddleware):
        @hook_config(can_jump_to=["model"])
        async def aafter_model(self, state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
            return None

    async_middleware = AsyncOnlyMiddleware()
    async_middleware.tools = []

    # Should return can_jump_to from overridden async method
    assert _get_can_jump_to(async_middleware, "after_model") == ["model"]


def test_async_middleware_with_can_jump_to_graph_snapshot(snapshot: SnapshotAssertion) -> None:
    """Test that async middleware with can_jump_to creates correct graph structure with conditional edges."""

    # Test 1: Async before_model with can_jump_to
    @before_model(can_jump_to=["end"])
    async def async_before_with_jump(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        if len(state["messages"]) > 5:
            return {"jump_to": "end"}
        return None

    agent_async_before = create_agent(
        model=FakeToolCallingModel(), middleware=[async_before_with_jump]
    )

    assert agent_async_before.get_graph().draw_mermaid() == snapshot

    # Test 2: Async after_model with can_jump_to
    @after_model(can_jump_to=["model", "end"])
    async def async_after_with_jump(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        if state["messages"][-1].content == "retry":
            return {"jump_to": "model"}
        return None

    agent_async_after = create_agent(
        model=FakeToolCallingModel(), middleware=[async_after_with_jump]
    )

    assert agent_async_after.get_graph().draw_mermaid() == snapshot

    # Test 3: Multiple async middleware with can_jump_to
    @before_model(can_jump_to=["end"])
    async def async_before_early_exit(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        return None

    @after_model(can_jump_to=["model"])
    async def async_after_retry(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        return None

    agent_multiple_async = create_agent(
        model=FakeToolCallingModel(),
        middleware=[async_before_early_exit, async_after_retry],
    )

    assert agent_multiple_async.get_graph().draw_mermaid() == snapshot

    # Test 4: Mixed sync and async middleware with can_jump_to
    @before_model(can_jump_to=["end"])
    def sync_before_with_jump(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        return None

    @after_model(can_jump_to=["model", "end"])
    async def async_after_with_jumps(state: AgentState, runtime: Runtime) -> dict[str, Any] | None:
        return None

    agent_mixed = create_agent(
        model=FakeToolCallingModel(),
        middleware=[sync_before_with_jump, async_after_with_jumps],
    )

    assert agent_mixed.get_graph().draw_mermaid() == snapshot


def test_dynamic_prompt_decorator() -> None:
    """Test dynamic_prompt decorator with basic usage."""

    @dynamic_prompt
    def my_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
        return "Dynamic test prompt"

    assert isinstance(my_prompt, AgentMiddleware)
    assert my_prompt.state_schema == AgentState
    assert my_prompt.tools == []
    assert my_prompt.__class__.__name__ == "my_prompt"

    # Verify it modifies the request correctly
    original_request = ModelRequest(
        model="test-model",
        system_prompt="Original",
        messages=[HumanMessage("Hello")],
        tool_choice=None,
        tools=[],
        response_format=None,
    )
    result = my_prompt.modify_model_request(
        original_request, {"messages": [HumanMessage("Hello")]}, None
    )
    assert result.system_prompt == "Dynamic test prompt"


def test_dynamic_prompt_uses_state() -> None:
    """Test that dynamic_prompt can use state information."""

    @dynamic_prompt
    def custom_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
        msg_count = len(state["messages"])
        return f"Prompt with {msg_count} messages"

    # Verify it uses state correctly
    original_request = ModelRequest(
        model="test-model",
        system_prompt="Original",
        messages=[HumanMessage("Hello")],
        tool_choice=None,
        tools=[],
        response_format=None,
    )
    result = custom_prompt.modify_model_request(
        original_request, {"messages": [HumanMessage("Hello"), HumanMessage("World")]}, None
    )
    assert result.system_prompt == "Prompt with 2 messages"


def test_dynamic_prompt_integration() -> None:
    """Test dynamic_prompt decorator in a full agent."""

    prompt_calls = 0

    @dynamic_prompt
    def context_aware_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
        nonlocal prompt_calls
        prompt_calls += 1
        return f"you are a helpful assistant."

    agent = create_agent(model=FakeToolCallingModel(), middleware=[context_aware_prompt])
    # Agent is already compiled

    result = agent.invoke({"messages": [HumanMessage("Hello")]})

    assert prompt_calls == 1
    assert result["messages"][-1].content == "you are a helpful assistant.-Hello"


async def test_async_dynamic_prompt_decorator() -> None:
    """Test dynamic_prompt decorator with async function."""

    @dynamic_prompt
    async def async_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
        return "Async dynamic prompt"

    assert isinstance(async_prompt, AgentMiddleware)
    assert async_prompt.state_schema == AgentState
    assert async_prompt.tools == []
    assert async_prompt.__class__.__name__ == "async_prompt"


async def test_async_dynamic_prompt_integration() -> None:
    """Test async dynamic_prompt decorator in a full agent."""

    prompt_calls = 0

    @dynamic_prompt
    async def async_context_prompt(
        request: ModelRequest, state: AgentState, runtime: Runtime
    ) -> str:
        nonlocal prompt_calls
        prompt_calls += 1
        return f"Async assistant."

    agent = create_agent(model=FakeToolCallingModel(), middleware=[async_context_prompt])
    # Agent is already compiled

    result = await agent.ainvoke({"messages": [HumanMessage("Hello")]})
    assert prompt_calls == 1
    assert result["messages"][-1].content == "Async assistant.-Hello"


def test_dynamic_prompt_overwrites_system_prompt() -> None:
    """Test that dynamic_prompt overwrites the original system_prompt."""

    @dynamic_prompt
    def override_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
        return "Overridden prompt."

    agent = create_agent(
        model=FakeToolCallingModel(),
        system_prompt="Original static prompt",
        middleware=[override_prompt],
    )
    # Agent is already compiled

    result = agent.invoke({"messages": [HumanMessage("Hello")]})
    assert result["messages"][-1].content == "Overridden prompt.-Hello"


def test_dynamic_prompt_multiple_in_sequence() -> None:
    """Test multiple dynamic_prompt decorators in sequence (last wins)."""

    @dynamic_prompt
    def first_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
        return "First prompt."

    @dynamic_prompt
    def second_prompt(request: ModelRequest, state: AgentState, runtime: Runtime) -> str:
        return "Second prompt."

    # When used together, the last middleware in the list should win
    # since they're both modify_model_request hooks executed in sequence
    agent = create_agent(model=FakeToolCallingModel(), middleware=[first_prompt, second_prompt])
    # Agent is already compiled

    result = agent.invoke({"messages": [HumanMessage("Hello")]})
    assert result["messages"][-1].content == "Second prompt.-Hello"
